#!/usr/bin/python
#$Revision: 1.24 $
import getopt,sys,re
import os
import Mass
import Utils
import Spectrum

try:
  from Numeric import average
except:
  print >>sys.stderr, "Could not import Numeric in Match. Some functions won't work."
  
class Match:

  def __init__(self):

    self.peptide = "No peptide"
    self.spectrum = None
    self.score = -1
    self.secondary_score = -1
    self.algorithm = "No algorithm"
    self.charge = -1
    self.spectrum_file = "Spectrum Id"

  # Eventually the firstScan, lastScan fields will come from 
  # an internal spectrum object.

  def __getitem__(self, key):
    return self.data[key]

  def __setitem__(self, key, item):
    self.data[key] = item    
  
  # TODO add a toString funciton



class SqtMatch(Match):
  "Contains information in a single SQT match"

  def __init__(self):
    Match.__init__(self)
    
    # TODO reconcile with Match object
    self.data = {}
    self.data["first scan"] = -1
    self.data["charge"] = -1
    self.data["peptide"] = ""
    

    # SqtMatch specific fields
    self.data["lastScan"] = -1
    self.data["time"] = -1
    self.data["server"] = -1
    self.data["M+H+"] = -1
    self.data["total intensity"] = -1
    self.data["low Sp"] = -1
    self.data["number seqs match"] = -1
    self.data["xcorr"] = -1
    self.data["deltCN"] = -1

    self.data["allLines"] = []
    self.data["locus"] = ""
    self.data["redundancy"] = ""
    self.data["id"] = ""
    self.data["is normal"] = ""

    # this field is kind of a hack. it is there to store matches
    # that are not the top match. really makes the SqtMatch contain multiple 
    # matches. Need to find a better way to do this. TODO
    self.data["submatches"] = []

  # Eventually the firstScan, lastScan fields will come from 
  # an internal spectrum object.
  def __getitem__(self, key):
    return self.data[key]

  def __setitem__(self, key, item):
    if key == "firstScan" or key == "lastScan":
      item = Utils.depad(item)
    if key == "peptide":
      self.peptide = str(item)
    self.data[key] = item    
  
  def __str__(self):
    firstLine = "\t".join(map(str,[ \
    "S",\
    self.data["firstScan"],\
    self.data["lastScan"],\
    self.data["charge"],\
    self.data["time"],\
    self.data["server"],\
    self.data["M+H+"],\
    self.data["total intensity"],\
    self.data["low Sp"],\
    self.data["number seqs match"]]))

    self["allLines"][0] = firstLine # FIX  
    return "".join( self["allLines"] )

  def calcDm(self):
    self["dM"] = abs(Mass.peptide_mass(self["peptide"], z=1) - self["M+H+"])

class ParseSpectrumException(Exception):
  def __init__(self, string="No string"):
    self.string = string
  def __str__(self):
    return self.string

class PepnovoParseSpectrumException(ParseSpectrumException):
  def __init__(self, string="No string"):
    self.string = string
  def __str__(self):
    return "ERROR: " + self.string

class PepnovoMatch(Match):
  "Contains information in a single Pepnovo match"

  def __init__(self, lines, specCollection):
    
    self.data = {}
    self.data["allLines"] = lines
    try:
      self.parseLines(lines, specCollection)
    except PepnovoParseSpectrumException, pe:
      raise pe

  def __setitem__(self, key, item):
    if key == "probs":
      self.data["confidence"] = average(item)
    self.data[key] = item    

  def __getitem__(self, key):
    return self.data[key]
    
  def parseLines(self, lines, specCollection):
    seen_first_digit = False
    seen_Start = False
    probs = []
    last_mass = 0.0
    last_amino = "X"
    for line in lines:
      if line.startswith("File "):
        fields = line.split()
        filename = fields[2]
        self.parseFilename(filename)
      elif line.startswith("Error"):
        raise PepnovoParseSpectrumException("Error in spectrum file")
      elif line.startswith("Start"):
        seen_Start = True
      elif line[0].isdigit() and seen_Start:
        if not seen_first_digit:
          # peptide line
          seen_first_digit = True 
          if line[0] != '0':
            # sequence does not start at beginning of spectrum
            raise PepnovoParseSpectrumException("Spectrum does not start with a zero")
          else:
            self.data["peptide"] = line.split()[1]
        else:
          fields = line.split()
          probs.append(float(fields[2]))
          last_mass = float(fields[0])
          last_amino = fields[1]

    # print >>sys.stderr, "Possibly broken method call below!"
    spectrum = specCollection.getSpectrum(\
        Spectrum.spectrumUID(self["firstScan"], os.path.basename(specCollection["filename"]), self["charge"]))
    if spectrum == None:
      raise SystemExit, "Could not find spectrum in specCollection!"
    # should be close to mass of spectrum
    real_mass = last_mass + Mass.mono_mass(last_amino) + Mass.mono_mass("H2O")
    if not ((spectrum["assumed m"] < real_mass + Mass.tol_mass("prec")) and \
        (real_mass - Mass.tol_mass("prec") < spectrum["assumed m"]) ):
      raise PepnovoParseSpectrumException("Mass off!")  
          
    self["probs"] = probs
    self["spectrum"] = spectrum 
  
  def parseFilename(self,filename):
    if filename.endswith(".dta"):
      fields = filename.split('.')
      self["filename"]  = fields[0]
      self["firstScan"] = int(fields[1])
      self["lastScan"] = fields[2]
      self["charge"] = fields[3]
    else:
      print >>sys.stderr, "PepnovoMatch can't deal with non dta filenames!"
      sys.exit(-1)

  def __str__(self):
    return '\t'.join(map(str, [ self["filename"], self["spectrum"]["first scan"], \
      "None", self["peptide"], self["charge"], self["confidence"]]\
      )) + '\n'
  
  
