#!/usr/bin/python
#-------------------
# $Revision: 1.58 $
#-------------------
import getopt,sys,re
from Match import *
from Spectrum import *
from SpectrumCollection import * 
import os

class MatchCollection:
  def __init__(self):
    self.data = {}
    self.data["file"] = ""
    self.data["fileName"] = ""
    self.data["matches"] = []
    self.data["matchDict"] = {}

  def __getitem__(self, key):
    return self.data[key]

  def __setitem__(self, key, item):
    self.data[key] = item    
  
  def appendMatch(self, m):
    self["matches"].append(m)
  
  def getMatches(self):
    return self["matches"]
 
  def getSpectrum(self, suid):
    for match in self["matches"]:
      if str(match.spectrum.uid()) == str(suid):
        return match.spectrum
    return None

  def getMatch(self, suid):
    import re
    for match in self["matches"]:
      if str(match.spectrum.uid()) == str(suid):
        return match
      fixed = re.sub(",", "-", match.spectrum.uid())
      if str(fixed) == str(suid):
        return match
    return None

class PsmMatchCollection(MatchCollection):
  """Holds the information in a PSM file"""

  def __init__(self, psmFileName=None, type="psm"):
    MatchCollection.__init__(self)
    self.data["file"] = psmFileName
    if type=="omics":
      self.parsePositivesFile(psmFileName)
    else:
      self.parse_psm_file(psmFileName)


  def __getitem__(self, key):
    return self.data[key]

  def __setitem__(self, key, item):
    self.data[key] = item    
 
  def addPsmsForSpectrum(self, spectrum, peps):
    for pep in peps:
      match = Match()
      match.spectrum = spectrum
      match.peptide = pep
      self.appendMatch(match)

  def allSpectra(self):
    allSpecs = []
    for match in self["matches"]:
      allSpecs.append(match.spectrum)
    return allSpecs

  def parsePositivesFile(self, psmFileName):
    """Parse the information in omics list of positives file"""
    if psmFileName == None: return
    file = open(psmFileName)
    seenFirst = False
    for line in file:
      if seenFirst == False: 
        seenFirst = True
        continue
      m = Match()
      fields = line.split()
      if len(fields) == 0:
        continue
      m.algorithm = "omics"
      dtaName = fields[0].lstrip('./')
      list = dtaName.split('.')
      groupId, firstScan = list[0], list[1]
      charge = fields[1]
      m.peptide = fields[3].split('.')[1]
      m.spectrumUID = spectrumUID(firstScan, groupId, charge)
      m.score  = '0'
      self.appendMatch(m)
    file.close()

  # FIX This routine is pretty gross. It would be nice to have one file format. Unfortunately,
  # I have about three or four, all of which are for PSMs. 
  # The preferred format is 
  # 
  #  Algorithm <\t> SpectrumUID <\t> Peptide <\t> Score <\n>
  # Which I will try to use from now on
  def parse_psm_file(self, psmFileName):
    """Parse the information in a PSM file"""
    if psmFileName == None: return
    file = open(psmFileName)
    for line in file:
      m = Match()
      fields = line.split("\t")
      if len(fields) > 4:
        m.algorithm = fields[0]
        m.spectrumUID = spectrumUID(fields[2], fields[1], fields[4])
# FIX believe below works only for dta
#      idsplit = fields[0].split('.')
#      idsplit[1] = idsplit[1].lstrip('/')
#      m.spectrumUID = spectrumUID(idsplit[2], idsplit[1], fields[1])
        m.peptide = fields[3]
      else: 
        m.algorithm = fields[0]
        m.spectrumUID = fields[1]
        m.peptide = fields[2]
        m.score = fields[3]
      if len(fields) > 5: m.score  = fields[5]
      else: m.score = 0.0
      if len(fields) > 6: m.secondary_score = fields[6]
      else: m.secondary_score =  0.0
      self.appendMatch(m)
    file.close()

  def addSpectrumInfo(self, spectrumCollection):
    """Attempt to add spectrum information to the matches in PsmMatchCollection
    using the information in multiSpectrumCollection"""

    for m in self["matches"]:
      suid = m.spectrumUID
      try:
        m.spectrum = spectrumCollection.getSpectrum(suid)
        if m.spectrum["parsePeaks"] == False: m.spectrum.parsePeaks()
        m.charge = m.spectrum["charge"]
      except:
        continue

class SqtMatchCollection(MatchCollection):
  "Parse out the information in an .sqt file."

  # Constructor takes a file as input.
  # TODO add constructor from MatchCollection object
  def __init__(self, saveSpace = False):
    MatchCollection.__init__(self)
    self.data["cachedLine"] = ""
    self.data["headerLines"] = []
    self.data["fileName"] = None
    self.saveSpace = saveSpace

  # TODO at some point have ability to say "scan number" and retrieve match
  # or "peptide" and retrieve list of matches
  def __getitem__(self, key):
    return self.data[key]

  def __setitem__(self, key, item):
    self.data[key] = item    

  def __str__(self):
    if len(self["matches"]) == 0:
      self.allMatches()

    outstr = "".join(self["headerLines"])
    outstr = outstr + "".join(map(str, self["matches"]))
    return outstr

  def parseFromFilename(self, filename):
    if not filename.endswith(".sqt"): 
      print >>sys.stderr, \
        "WARNING: Filename for SqtMatchCollection doesn't end with '.sqt'."
    self.data["fileName"]  = filename
    self.data["file"] = open(filename, 'r')

  def outputToFilename(self, filename):
    # if filename == file: filename = self.data["file"] 
    # TODO  check for same name as parent

    if not filename.endswith(".sqt"): 
      print >>sys.stderr, \
        "ERROR: Output filename for SqtMatchCollection must end with '.sqt'."
      sys.exit(-1)

    file  = open(filename, 'w')
    # str will read in all lines in file if not already
    file.write(str(self))
    self["file"]=file
    file.close()

  def parseFirstLine(self, m, l):
    results = re.split("\t", l)
    m["firstScan"] = results[1]
    m["lastScan"]  = results[2]
    m["charge"]    = int(results[3])
    m["time"]  = results[4]
    m["server"]  = results[5]
    m["M+H+"]  = float(results[6])
    m["total intensity"]  = results[7]
    m["low Sp"]  = results[8]
    m["number seqs match"]  = results[9]
    groupId =  os.path.basename(self["fileName"]).split('.')[0]
    m.spectrumUID = spectrumUID(m["firstScan"], groupId, m["charge"])

  def parseMatchInfo(self, match, lines):
    #print lines
    #print "\n\n======\n\n"

    mlines = filter(lambda x: x[0]=='M', lines)
    if len(mlines) == 0:
      print lines
    first_m_line = mlines[0]
    l_lines = filter(lambda x: x[0]=='L', lines)
    first_l_line = l_lines[0]
    if len(mlines) > 1:
      second_m_line = mlines[1]
    else:
      second_m_line = ""
    if first_m_line[0:2] != "M\t" or first_l_line[0:2] != "L\t":
      print >>sys.stderr, "Problem with sqt file line!", lines
      sys.exit(1)

    results = re.split("\t", first_m_line)
    match["xcorr"] = results[5]
    match["sp"] = results[6]
    match["rSp"] = results[2]
    match["matched ions"] = results[7]
    match["expected ions"] = results[8]
    try:
      left, peptide, right = results[9].split('.')
    except:
      raise SystemExit, "Problem parsing %s : %s" % (first_m_line, results[9])
    match["left"] = re.sub(' ', '', left)
    match["peptide"] = peptide
    match["right"] = re.sub(' ','', right)
    match.calcDm()
    if second_m_line[0:2] == "M\t": 
      results = re.split("\t", second_m_line)
      match["deltCN"] = results[4]
    else:
      # If there is no other hit set the deltCN to the xcorr value
      match["deltCN"] = match["xcorr"] 
    results = re.split("\t", first_l_line)
    match["locus"] = results[1]

    if self.saveSpace == False:
      for m_line in mlines:
        subMatch = SqtMatch()
        results = re.split("\t", m_line)
        subMatch["xcorr"] = results[5]
        subMatch["sp"] = results[6]
        subMatch["rSp"] = results[2]
        subMatch["matched ions"] = results[7]
        subMatch["expected ions"] = results[8]
        left, peptide, right = results[9].split('.')
        subMatch["left"] = re.sub(' ','', left)
        subMatch["peptide"] = peptide
        subMatch["right"] = re.sub(' ','', right)
        subMatch["deltCN"]  = results[5]
        subMatch.calcDm()
        match["submatches"].append(subMatch)

  def nextMatch(self):

    if self["cachedLine"] == "":
      # parse header
      l = self["file"].readline()
      if l == "": return None
      while l[0:2] != "S\t":
        self["headerLines"].append(l) # FIX don't think this is working
        l = self["file"].readline()
        if l == "": return None
      self["cachedLine"] = l

    lines = []
    lines.append(self["cachedLine"])

    l = self["file"].readline()
    if l == "": return None

    while l[0:2] != "S\t":
      lines.append(l)
      l = self["file"].readline()
      if l == "": break

    # could be dangerous recursion
    if len(lines) < 4: return self.nextMatch() 
    
    self["cachedLine"] = l
  
    match = SqtMatch()
    self.parseFirstLine(match, lines[0])
    self.parseMatchInfo(match, lines)

    # is this match a normal match? TODO - move to a separate routine
    match["is normal"] = False
    for line in lines:
      if not line.startswith('L'): continue
      if not 'rand' in line: 
        match["is normal"] = True
        break

    if self.saveSpace == False: match["allLines"] = lines

    self.appendMatch(match)

    return match  
  
  def appendMatch(self, match):
    self["matches"].append(match)
    # TODO create a matchUID as well
    key = match["firstScan"], int(match["charge"])

    if self["matchDict"].has_key(key):
      print >>sys.stderr, \
        "MatchCollection keys (firstScan, charge) must be unique."
      exit(-1)

    self["matchDict"]["key"] = match

  def allMatches(self):
    "Returns a list of all the Matches in this collection."

    if len(self["matches"])==0:
      # we must not have read in the matches yet
      match = self.nextMatch()
      while match != None:
        # match appends itself to the match list
        match = self.nextMatch()
    
    return self["matches"]

  def getMatch(self, suid):
    for match in self["matches"]:
      if (int(match["firstScan"]) == int(uidSpectrumID(suid))):
        return match
    return None

  def addSpectrumInfo(self, spectrumCollection):
    """Attempt to add spectrum information to the matches in PsmMatchCollection
    using the information in multiSpectrumCollection"""

    for m in self["matches"]:
      suid = m.spectrumUID
      try:
        m.spectrum = spectrumCollection.getSpectrum(suid)
        m.charge = m.spectrum["charge"]
      except Exception, e:
        print >>sys.stderr, "Trouble with %s %s " % (suid, e)
        continue

#----
# Main
#----
if __name__ == '__main__':
  if len(sys.argv) < 2:
    print "Usage: %s <sqt-filename>" % (sys.argv[0])
    sys.exit(1)
 
  for idx in range(1,len(sys.argv)): 
    sqt = SqtMatchCollection()
    sqt.parseFromFilename(sys.argv[idx])
    matches = sqt.allMatches()
    step = sys.argv[idx].split("-")[-1].split(".")[0]
    for match in matches:
      # if int(match["charge"]) != 1: continue
      # print "%s\t%s\t%s" % \
      #  (step, match["firstScan"], match["peptide"])
      x = float(match["xcorr"])
      d = float(match["deltCN"])
      uid  = match["firstScan"] + "," + sys.argv[idx] + "," + str(match["charge"])
      print "%s\t%.5f\t%.5f" % (uid, x, x - d*x)
