import Numeric
import random
import math
import MLab
import os
import tempfile
import copy
import time

import myio,misc

'''functionality for assessing classifier performance'''

__docformat__ = "restructuredtext en"


def test(classifier, data, **args) :
    '''test a classifier on a given dataset
    
    :Parameters:
      - `classifier` - a trained classifier
      - `data` - a dataset
      
    :Return:
      a Results class instance

    :Keywords:
      - `stats` - whether to compute the statistics of the match between the
        predicted labels and the given labels [True by default]
   	'''

    if 'verbose' in args :
        verbose = args['verbose']
    else :
        verbose = 1
        
    if verbose :
        print 'testing', \
              '***********************************************************'

    testStart = time.clock()

    if data.testingFunc is not None :
        data.test(classifier.trainingData, **args)

    classifier.preproject(data)

    res = classifier.resultsObject(data, classifier, **args)

    for i in range(len(data)) :
        if verbose and i % 100 == 0 and i > 0 :
            print i, 'patterns classified'
        res.appendPrediction(classifier.classify(data, i), data, i)

    try :
	res[0].log = classifier.log
    except :
	pass
    try :
        computeStats = args['stats']
    except :
        computeStats = False
    if computeStats and data.labels.L is not None :
        res.computeStats()

    try :
	res[0].log.testingTime = time.clock() - testStart
    except :
	pass

    return res


def loo(classifier, data, **args) :
    '''perform Leave One Out Cross Validation
    USAGE: loo(classifier, data)
    returns a Results structure; results represent the average over the folds
    the ROC curve is computed using the resulting classification of each
    point in the dataset (in contrast to Provost, Fawcett and Kohavi who compute
    average ROC curves).
    '''

    looResults = Results()
    args['stats'] = False

    for i in range(len(data)) :

        trainingPatterns = misc.setminus(range(len(data)), [i])
        looResults.append(
            classifier.trainTest(data, trainingPatterns, [i], **args))

    looResults.computeStats()
   
    return looResults


def cvFromFolds(classifier, data, trainingPatterns, testingPatterns,
                **args) :
    
    """perform cross validation

    :Parameters:
      - `classifier` - a classifier template
      - `data` - a dataset
      - `trainingPatterns` - a list providing the training examples for each fold
      - `testingPatterns` - a list providing the testing examples for each fold

    :Keywords:
      - `intermediateFile` - a file name to save intermediate results under
        if this argument is not given, no intermediate results are saved

    :Returns:
      a Results object. Results represent the average over the folds.
      The ROC curve is computed using the resulting classification of each
      point in the dataset (in contrast to Provost, Fawcett and Kohavi who compute
      average ROC curves).
    """

    assert len(trainingPatterns) == len(testingPatterns)

    cvResults = Results()
    args['stats'] = False
    
    for fold in range(len(trainingPatterns)) :
        cvResults.append(
            trainTest(classifier, data,
                      trainingPatterns[fold], testingPatterns[fold], **args))
        #if 'intermediateFile' in args :
        #    cvResults.save(args['intermediateFile'])

    cvResults.computeStats()
        
    return cvResults


def cv(classifier, data, numFolds = 5, **args) :
    """perform k-fold cross validation

    :Parameters:
      - `classifier` - a classifier template
      - `data` - a dataset
      - `numFolds` - number of cross validation folds (default = 5)

    :Returns:
      a Results object. Results represent the average over the folds.
      The ROC curve is computed using the resulting classification of each
      point in the dataset (in contrast to Provost, Fawcett and Kohavi who compute
      average ROC curves).

    :Keywords:
      - `numFolds` - number of cross validation folds (default = 5)
      - `seed` - random number generator seed
    """

    if 'numFolds' in args :
        numFolds = args['numFolds']
    if 'seed' in args :
        random.seed(args['seed'])
    perm = range(len(data))
    random.shuffle(perm)
    foldSize = len(data) / numFolds
    trainingPatterns = []
    testingPatterns = []

    for fold in range(numFolds) :
        if fold < numFolds-1:
            testingPatterns.append(perm[foldSize * fold : foldSize * (fold + 1)])
        else:
            testingPatterns.append(perm[foldSize * fold : len(data)])
        trainingPatterns.append(misc.setminus(range(len(data)),
                                              testingPatterns[-1]))
        
    return cvFromFolds(classifier, data, trainingPatterns, testingPatterns, **args)

        
def stratifiedCV(classifier, data, numFolds = 5, **args) :
    """perform k-fold stratified cross-validation; in each fold the number of
    patterns from each class is proportional to the relative fraction of the
    class in the dataset

    :Parameters:
      - `classifier` - a classifier template
      - `data` - a dataset
      - `numFolds` - number of cross validation folds (default = 5)

    :Returns:
      a Results object. Results represent the average over the folds.
      The ROC curve is computed using the resulting classification of each
      point in the dataset (in contrast to Provost, Fawcett and Kohavi who compute
      average ROC curves).

    :Keywords:
      - `numFolds` - number of cross-validation folds -- overrides the numFolds parameter
      - `seed` - random number generator seed
      - `trainingAllFolds` - a list of patterns that are to be used as training
        examples in all CV folds.
      - `intermediateFile` - a file name to save intermediate results under
        if this argument is not given, not intermediate results are saved
    """

    if 'numFolds' in args :
	numFolds = args['numFolds']
    if 'seed' in args :
        random.seed(args['seed'])
    if 'trainingAllFolds' in args :
        trainingAllFolds = args['trainingAllFolds']
    else :
        trainingAllFolds = []
    trainingAllFoldsDict = misc.list2dict(trainingAllFolds)

    labels = data.labels
    p = [[] for i in range(labels.numClasses)] 
    classFoldSize = Numeric.array([0] * labels.numClasses)
    for k in range(labels.numClasses):
        classFoldSize[k] = int(labels.classSize[k] / numFolds)
    for i in range(len(data)):
        if i not in trainingAllFoldsDict :
            p[labels.Y[i]].append(i)
    for k in range(labels.numClasses):
        random.shuffle(p[k])

    trainingPatterns = [[] for i in range(numFolds)]
    testingPatterns = [[] for i in range(numFolds)]
    for fold in range(numFolds) :
        for k in range(labels.numClasses) :
            classFoldStart = classFoldSize[k] * fold
            if fold < numFolds-1:
                classFoldEnd = classFoldSize[k] * (fold + 1)
            else:
                classFoldEnd = labels.classSize[k]
            testingPatterns[fold].extend(p[k][classFoldStart:classFoldEnd])
            if fold > 0:
                trainingPatterns[fold].extend(p[k][0:classFoldStart] +
                                              p[k][classFoldEnd:labels.classSize[k]])
            else:
                trainingPatterns[fold].extend(p[k][classFoldEnd:labels.classSize[k]])

    if len(trainingPatterns) > 0 :
        for fold in range(len(trainingPatterns)) :
            trainingPatterns[fold].extend(trainingAllFolds)
        
    return cvFromFolds(classifier, data, trainingPatterns, testingPatterns, **args)


def nCV(classifier, data, **args) :
    """
    runs CV n times, returning a 'ResultsList' object.

    :Parameters:
      - `classifier` - classifier template
      - `data` - dataset
      
    :Keywords:
      - `cvType` - which CV function to apply (default: stratifiedCV)
      - `seed` - random number generator seed (default: 1)
        This is used as the seed for the first CV run.  Subsequent runs
        use seed + 1, seed + 2...
      - `iterations` - number of times to run CV (default: 10)
      - `numFolds` - number of folds to use with CV (default: 5)
      - `intermediateFile` - a file name to save intermediate results under
        if this argument is not given, no intermediate results are saved

    :Returns:
      `ResultsList` - a list of the results of each CV run as a ResultsList object
    """

    cvList = ResultsList()

    cvType = 'stratifiedCV'
    if 'cvType' in args : cvType = args['cvType']
    seed = 1
    if 'seed' in args : seed = args['seed']
    numFolds = 5
    if 'numFolds' in args : numFolds = args['numFolds']
    iterations = 10
    if 'iterations' in args : iterations = args['iterations']
    intermediateFile = None
    if 'intermediateFile' in args : intermediateFile = args['intermediateFile']

    for i in range(iterations) :
        if cvType == 'stratifiedCV' :
            cvList.append(stratifiedCV(
                classifier, data, numFolds, seed = seed + i))
        elif cvType == 'cv' :
            cvList.append(cv(classifier, data, numFolds, seed = seed + i))
        else :
            raise ValueError, 'unrecognized type of CV'
        if intermediateFile is not None :
            cvList.save(intermediateFile)

    cvList.computeStats()

    return cvList
    

def makeFolds(data, numFolds, datasetName, directory = '.') :

    '''split a dataset into several folds and save the training and testing
    data of each fold as a separate dataset

    data - a dataset instance
    numfolds - number of folds into which to split the data
    datasetName - string to use for the file names
    directory - the directory into which to deposit the files
    '''
        
    perm = range(len(data))
    random.shuffle(perm)
    foldSize = len(data) / numFolds
    
    for fold in range(numFolds) :
        if fold < numFolds-1:
            testingPatterns = perm[foldSize * fold : foldSize * (fold + 1)]
        else:
            testingPatterns = perm[foldSize * fold : len(data)]
        trainingPatterns = misc.setminus(range(len(data)), testingPatterns)

        trainingData = data.__class__(data, patterns = trainingPatterns)
        testingData = data.__class__(data, patterns = testingPatterns)

        testingDataName = directory + '/' + datasetName + 'Testing' + str(fold) + '.data'
        testingData.save(testingDataName)
        trainingDataName = directory + '/' + datasetName + 'Training' + str(fold) + '.data'
        trainingData.save(trainingDataName)
    

def cvFromFile(classifier, trainingBase, testingBase, datasetClass, **args) :
    '''perform CV when the training and test data are in files whose names
    are of the form:
    trainingBase + number + string
    and
    testingBase + number + string
    For example:
    training0.data, training1.data, training2.data
    and
    testing0.data, testing1.data, testing2.data
    for 3 fold CV.
    training and testing files are matched by the number appearing after
    the strings trainingBase and testingBase
    both trainingBase and testingBase can be paths.
    '''

    args['stats'] = False
    import re
    directory = os.path.dirname(trainingBase)
    if directory == '' : directory = '.'
                
    files = os.listdir(directory)
    trainingFiles = [file for file in files
                     if file.find(trainingBase) == 0]
    testingFiles = [file for file in files
                    if file.find(testingBase) == 0]

    # now we check if the training files match the test files:
    numberRE = re.compile(r'\d+')

    trainingNum = [numberRE.findall(trainingFile)[-1]
                   for trainingFile in trainingFiles]
    testingNum = [numberRE.findall(testingFile)[-1]
                  for testingFile in testingFiles]    

    assert len(trainingNum) == len(testingNum)
    for i in range(len(trainingNum)) :
        if trainingNum[i] != testingNum[i] :
            raise ValueError, 'training files do not match testing files'

    print 'trainingFiles',trainingFiles

    trainingData = datasetClass(trainingFiles[0])

    cvResults = classifier.resultsObject(trainingData, classifier)

    for fold in range(len(trainingFiles)) :
        if fold > 0 :
	    trainingData = datasetClass(trainingFiles[fold])

        classifier.train(trainingData)
        del trainingData
        
	testingData = datasetClass(testingFiles[fold])

        r = classifier.test(testingData, **args)
        cvResults.extend(r)

    cvResults.computeStats()
        
    return cvResults


def scatter(r1, r2, statistic = 'roc', x1Label = '', x2Label= '',
            fileName = None, **args) :
    """
    a scatter plot for comparing the performance of two classifiers

    :Parameters:
      - `r1, r2` - both are either a list of Result classes, or a list of
        success rates / ROC scores
      - `statistic` - which measure of classifier success to plot
        values : 'roc', 'successRate', 'balancedSuccessRate'
	in order to specify parts of the roc curve you can use something like:
	'roc50' or 'roc0.1'

    :Keywords:
      - `title` - the title of the plot
    """

    if len(r1) != len(r2) :
        print 'unequal lengths for r1 and r2'
        if type(r1) != type({}) :
            raise ValueError, 'Cannot handle unequal length when it is not a dict'
        keys1 = r1.keys()
        keys2 = r2.keys()
        common = misc.intersect(keys1, keys2)
        r1new = {}
        r2new = {}
        for key in common :
            r1new[key] = r1[key]
            r2new[key] = r2[key]
        r1 = r1new
        r2 = r2new
        
    if type(r1) == type({}) and type(r2) == type({}) :
        I = r1.keys()
    else :
        I = range(len(r1))

    if (r1[I[0]].__class__.__name__ == 'Results' or
        r1[I[0]].__class__.__name__ == 'Container') :
        p1 = misc.extractAttribute(r1, statistic)
        p2 = misc.extractAttribute(r2, statistic)
    else :
        p1 = r1
        p2 = r2
        
    if type(p1) == type({}) :
        p1 = p1.values()
        p2 = p2.values()

    import matplotlib
    if fileName is not None and fileName.find('.eps') > 0 :
        matplotlib.use('PS')
    from matplotlib import pylab

    x = Numeric.arange(0,1,0.01)
    pylab.plot(p1, p2, 'bo',x,x, '-k')
    pylab.xlabel(x1Label, fontsize = 18)
    pylab.ylabel(x2Label, fontsize = 18)
    #title = {'roc50Area':'Area under ROC 50 curve',
    #         'rocArea' : 'Area under ROC curve',
    #         'successRate' : 'Success Rate',
    #         'balancedSuccessRate' : 'Balanced Success Rate'}
    if 'title' in args :
        #pylab.title(title[statistic], fontsize = 18)
        pylab.title(args['title'], fontsize = 18)
    pylab.show()

    if fileName is not None :
        pylab.savefig(fileName)
	pylab.close()

def plotROC2(decisionFunc, givenY, fileName = None, **args) :

    res = misc.Container({'decisionFunc': decisionFunc,
                          'givenY' : givenY,
                          'Y' : None})
    plotROC(res, fileName, **args)


def plotROC(res, fileName = None, **args) :
    """plot the ROC curve from a given Results (or Results-like) object

    :Parameters:
      - `res` - Results (or Container object that was made by saving a a
        Results object (note that if you have a Results object you can
        use this function as a method so there is no need to supply this
        argument).
      - `fileName` - optional argument - if given, the roc curve is saved
        in the given file name.  The format is determined by the extension.
        Supported extensions: .eps, .png, .svg
    
    :Keywords:
      - `rocN` - what type of ROC curve to plot (roc50, roc10 etc.) default is
        full ROC curve
      - `normalize` - whether to normalize the ROC curve (default: True)
      - `plotStr` - which string to pass to matplotlib's plot function
        default: 'ob'
      - `axis` - redefine the figure axes; takes a list of the form
        [xmin,xmax,ymin,ymax]
      - `show` - whether to show the ROC curve (default: True)
        useful when you just want to save the curve to a file.
        The use of Some file formats automatically sets this to False
        (e.g. svg files).  This relates to quirks of matplotlib.
    """

    if 'rocN' in args :
        rocN = args['rocN']
    else :
        rocN = None
    if 'show' in args :
        show = args['show']
    else :
        show = True
    if 'plotStr' in args :
        plotStr = args['plotStr']
    else :
        plotStr = 'ob'
    rocNormalize = True
    if 'normalize' in args :
        rocNormalize = args['normalize']

    numPoints = 200
    if 'numPoints' in args :
        numPoints = args['numPoints']
        
    targetClass = 1

    if type(res) == type([]) :
        feature = res[0]
        givenY = res[1]
        rocTP, rocFP, rocArea = roc(
            None, givenY, feature, rocN, targetClass, rocNormalize)
    else :
        rocTP, rocFP, rocArea = roc(
            res.Y, res.givenY, res.decisionFunc, rocN, targetClass,
            rocNormalize)    

    stride = int(max(1, float(len(rocTP)) / float(numPoints)))

    if stride > 1 :
        rocTP = [rocTP[i] for i in range(0,len(rocTP), stride)]
        rocFP = [rocFP[i] for i in range(0,len(rocFP), stride)]        
        
    import matplotlib
    if fileName is not None and fileName.find('.svg') > 0 :
        matplotlib.use('SVG')
        show = False
    if fileName is not None and fileName.find('.eps') > 0 :
        matplotlib.use('PS')
        show = False

    from matplotlib import pylab
    lines = pylab.plot(rocFP, rocTP, plotStr,
                        markersize = 8, linewidth = 3)
    if rocNormalize :
        pylab.xlabel('False positive rate', fontsize = 18)
        pylab.ylabel('True positive rate', fontsize = 18)
    else :
        pylab.xlabel('False positives', fontsize = 18)
        pylab.ylabel('True positives', fontsize = 18)
    if rocNormalize :
        pylab.axis([0, 1, 0, 1])
    if 'axis' in args :
        pylab.axis(args['axis'])
    print fileName
    if fileName is not None :
        pylab.savefig(fileName)
    if show :
        pylab.show()

def plotROCs(resList, descriptions = None, fileName = None, **args) :

    """
    plot multiple ROC curves.

    :Parameters:
      - `resList` - a list or dictionary of Result or Result-like objects
      - `descriptions` - text for the legend (a list the size of resList).
        A legend is not shown if this parameter is not given
        In the case of a dictionary input the description for the legend is
        taken from the dictionary keys.
      - `fileName` - if given, a file to save the figure in

    :Keywords:
      - `legendLoc` - the position of the legend -- an integer between 0 and 9;
        see the matplotlib documentation for details
      - `plotStrings` - a list of matlab style plotting string to send to the
        plotROC function (instead of the plotString keyword of plotROC)
      - `other keywords` - keywords of the plotROC function 
    """

    if type(resList) == type([]) and type(resList[0]) == type('') :
        fileNames = resList
        resList = []
        for fileName in fileNames :
            resList.append(myio.load(fileName))
        if descriptions is None :
            descriptions = []
            for fileName in fileNames :
                descriptions.append(os.path.splitext(fileName)[0])

    import matplotlib
    show = True
    if fileName is not None and fileName.find('.svg') > 0 :
        matplotlib.use('SVG')
        show = False
    if fileName is not None and fileName.find('.eps') > 0 :
        matplotlib.use('PS')
        show = False
                
    from matplotlib import pylab
    args['show'] = False
    
    plotStrings = ['bo', 'k^', 'rv', 'g<', 'm>', 'k<']
    plotStrings = ['b-', 'k--', 'r-', 'g-.', 'm-', 'k:', 'b-', 'r-', 'g-']
    #plotStrings = ['b:', 'k-.', 'b-', 'g-', 'm-', 'k-', 'b-', 'r-', 'g-']
    if 'plotStrings' in args :
        plotStrings = args['plotStrings']
    if type(resList) == type([]) :
        for i in range(len(resList)) :
            print i
            args['plotStr'] = plotStrings[i]
            plotROC(resList[i], **args)
    else :
        if descriptions is None :
            descriptions = [key for key in resList]
        i = 0
        for key in resList :
            args['plotStr'] = plotStrings[i]
            plotROC(resList[key], **args)
            i+=1

    if descriptions is not None :
        legendLoc = 'best'
        if 'legendLoc' in args :
            legendLoc = args['legendLoc']
        pylab.legend(descriptions, loc = legendLoc)
        leg = pylab.gca().get_legend()
        ltext  = leg.get_texts()  # all the text.Text instance in the legend
        llines = leg.get_lines()  # all the lines.Line2D instance in the legend
        frame  = leg.get_frame()  # the Rectangle instance surrounding the legend

        #frame.set_facecolor(0.80)     # set the frame face color to light gray
        for obj in ltext :
            obj.set_size(14)
        #leg.draw_frame(False)         # don't draw the legend frame

    print fileName
    
    if fileName is not None :
        pylab.savefig(fileName)
    if show :
        pylab.show()

    
def significance(r1, r2, statistic = 'roc') :
    """
    report the statistical significance of the difference in error rates
    of a series of classification results of two classifiers
    using the Wilcoxon signed rank test

    :Parameters:
      - `r1, r2` - both are either a list of Result classes, or a list of success
        rates
      - `statistic` - which measure of classifier success to plot
        values : 'roc', 'successRate', 'balancedSuccessRate'
	in order to specify parts of the roc curve you can use something like:
	'roc50' or 'roc0.1'

    """

    if len(r1) != len(r2) :
        print 'unequal lengths for r1 and r2'
        if type(r1) != type({}) :
            raise ValueError, 'Cannot handle unequal length when it is not a dict'
        keys1 = r1.keys()
        keys2 = r2.keys()
        common = misc.intersect(keys1, keys2)
        r1new = {}
        r2new = {}
        for key in common :
            r1new[key] = r1[key]
            r2new[key] = r2[key]
        r1 = r1new
        r2 = r2new

    if type(r1) == type({}) and type(r2) == type({}) :
        I = r1.keys()
    else :
        I = range(len(r1))
    if r1[I[0]].__class__.__name__ == 'Results' or r1[I[0]].__class__.__name__ == 'Container' :
        p1 = misc.extractAttribute(r1, statistic)
        p2 = misc.extractAttribute(r2, statistic)
    else :
        p1 = r1
        p2 = r2

    if type(p1) == type({}) :
        p1 = p1.values()
        p2 = p2.values()

        
    #import stats
    
    #p = stats.ranksums(p1, p2)[1]

    #print 'median of first:', MLab.median(p1), 'median of second:', MLab.median(p2)
    #print 'average of first:', MLab.mean(p1), 'average of second:', MLab.mean(p2)
    
    import salstat_stats
    test = salstat_stats.TwoSampleTests(p1, p2)
    test.SignedRanks (p1, p2)

    p = test.prob
    
    return test
    

def trainTest(classifierTemplate, data, trainingPatterns, testingPatterns, *options,
              **args) :
    '''Train a classifier on the list of training patterns, and test it 
    on the test patterns
    '''

    if 'verbose' in args :
        verbose = args['verbose']
    else :
        verbose = True

    trainingData = data.__class__(data, deepcopy = classifierTemplate.deepcopy,
                                  patterns = trainingPatterns)

    classifier = classifierTemplate.__class__(classifierTemplate)

    if verbose :
        print 'training', \
              '***********************************************************'

    classifier.train(trainingData, **args)

    testingData = data.__class__(data, deepcopy = True,
                                 patterns = testingPatterns)

    return classifier.test(testingData, **args)

#return subClassifier.test(data, testingPatterns, **args)


def confmat(L1, L2) :
    """computes the confusion matrix between two labelings
    """
    
    if len(L1) != len(L2):
        raise ValueError, "labels not the same length"

    n = len(L1)

    classes1 = misc.unique(L1)
    classes2 = misc.unique(L2)
    classes1.sort()
    classes2.sort()
    numClasses1 = len(classes1)
    numClasses2 = len(classes2)

    I1 = {}  # a mapping from classes1 to 0..numclasses1-1
    I2 = {}
    for i in range(numClasses1) :
        I1[classes1[i]] = i
    for i in range(numClasses2) :
        I2[classes2[i]] = i
    
    confmat = Numeric.zeros((numClasses1, numClasses2))

    for i in range(n):
        confmat[I1[L1[i]]][I2[L2[i]]] += 1
        
    return confmat


def superConfmat(Y1, Y2, numClasses = 0) :
    """computes the confusion matrix between two labelings, where
    the matrix is assumed to be square, according to the labels of L1
    L1 and L2 are assumed to have integer components in the range
    0,.., numClasses
    """

    if len(Y1) != len(Y2):
        raise ValueError, "labels not the same length"

    n = len(Y1)

    m = max(max(Y1), max(Y2), numClasses)

    # prefer using a list object rather Numeric object so that there
    # wouldn't be a problem in pickling the object
    confmat = misc.matrix((m, m), 0)

    for i in range(n):
        confmat[Y1[i]][Y2[i]] += 1
        
    return confmat


def roc(Y, givenY, decisionFunc, n = None, targetClass = 1, normalize = True) :
    """Compute the ROC curve and area under the curve for a two class problem

	:Parameters:
      - `Y` - the predicted labels (can put None instead)
      - `givenY` - the true labels
	  - `decisionFunc` - the values of the decision function
	  - `n` - the number of false positives to take into account (roc_n)
	  - `targetClass - the "positive" class
      - `normalize` whether to normalize the roc curve (default: True)
	     when this is set to False, TP/FP counts are output rather than TP/FP rates

    :Keywords:
            
        """
    
    if n is not None and n < 1 :
        n = int(n * Numeric.sum(Numeric.not_equal(givenY, targetClass)))
    I = range(len(decisionFunc))
    random.shuffle(I)
    decisionFunc = [decisionFunc[i] for i in I]
    givenY = [givenY[i] for i in I]
    f = Numeric.array(decisionFunc)

    tp = [0.0]
    fp = [0.0]
    I = Numeric.argsort(-f)
    
    for patternIdx in I :
        if givenY[patternIdx] == targetClass :
            tp[-1] += 1
        else :
            tp.append(tp[-1])
            fp.append(fp[-1] + 1.0)
        if n is not None and fp[-1] >= n :
            break

    numTP = Numeric.sum(Numeric.equal(givenY, targetClass))
    
    if normalize : 
        for i in range(len(tp)):
            #if tp[-1] > 0 : tp[i] /= float(tp[-1])
            if tp[-1] > 0 : tp[i] /= float(numTP)
        for i in range(len(fp)) :
            if fp[-1] > 0 : fp[i] /= float(fp[-1])

        area = Numeric.sum(tp) / len(tp)

    else :
        area = Numeric.sum(tp) / (len(tp) * numTP)

    return tp,fp, area


def rfp(res, Y) :
    '''Compute the median RFP rate :
    the RFP rate is defined for a positive example as the fraction of
    negative examples that have a higher score (decision function value)
    Input:
    res - a Results instance
    Y - given labels
    OPTIONS (from the Results object) :
    '''

    negativePatterns = Numeric.nonzero(Numeric.equal(Y, 0))
    positivePatterns = Numeric.nonzero(Numeric.equal(Y, 1))

    f = res.decisionFunc

    fneg = Numeric.take(f, negativePatterns)
    fneg = Numeric.sort(fneg)

    RFP = []
    RFP50 = []
    
    numNeg = float(len(fneg))
    for p in positivePatterns :
        numfp = numNeg - Numeric.searchsorted(fneg, f[p])
        RFP.append(numfp)
        RFP50.append(min(numfp,50))

    if len(RFP) > 0 :
        return MLab.median(RFP) / numNeg, MLab.median(RFP50) / 50.0
    else :
        return None, None


class ClassificationResults (object) :

    def __repr__(self) :

        rep = []
        if not hasattr(self, 'confusionMatrix') :
            return ''

        if self.numClasses == 1 : return ''
        
        rep.extend( self.formatConfusionMatrix() )

        rep.append('success rate: %f' % self.successRate)
        rep.append('balanced success rate: %f' % self.balancedSuccessRate)
        if self.numClasses == 2 :
            rep.append('ppv: %f ' % self.ppv + 'sensitivity: %f' % self.sensitivity)

	if self.numClasses == 2 :
	    rep.append('area under ROC curve: %f' % self.roc)
	    if int(self.rocN) == self.rocN :
		rep.append('area under ROC %d curve: %f ' % \
			   (self.rocN, getattr(self, 'roc' + str(self.rocN))) )
	    else :
		rep.append('area under ROC %f curve: %f ' % \
			   (self.rocN, getattr(self, 'roc' + str(self.rocN))) )

        return '\n'.join(rep)


    def formatConfusionMatrix(self) :

        rep = []
        columnWidth = 4
        for label in self.classLabels :
            if len(label) > columnWidth : columnWidth = len(label)
        columnWidth +=1
        columnWidth = int(max(
            columnWidth,
            math.ceil(math.log10(MLab.max(MLab.max(self.confusionMatrix))))+1))

        rep.append('Confusion Matrix:')
        rep.append(' ' * columnWidth + ' Given labels:')
        rep.append(' ' * columnWidth + 
		   ''.join([label.center(columnWidth) for label in self.classLabels]))

        (numClassesTest, numClassesTrain) = Numeric.shape(self.confusionMatrix)
        for i in range(numClassesTest) :
            label = self.classLabels[i]
            rep.append(label.rjust(columnWidth) + ''.join(
		    [str(self.confusionMatrix[i][j]).center(columnWidth)
		     for j in range(numClassesTrain)]))

        return rep

    def computeStats(self, **args) :

        if len(self.givenY) == 0 : return
        Y = self.givenY
        self.confusionMatrix = superConfmat(self.Y, self.givenY,
                                               self.numClasses)

        if hasattr (self, 'patternID') and self.patternID != [] :
            self.misclassified = misc.subList(self.patternID, Numeric.nonzero(
                Numeric.not_equal(self.Y, Y)))
        else :
            self.misclassified = Numeric.nonzero(Numeric.not_equal(self.Y, Y))

        self.successRate, self.balancedSuccessRate, self.ppv, \
                          self.sensitivity = self.successRates()

    def successRates(self) :

        targetClass = 1
        classSuccess = Numeric.zeros(self.numClasses, Numeric.Float)
        classSize = Numeric.zeros(self.numClasses, Numeric.Int)
        for i in range(len(self)) :
            classSize[self.givenY[i]] += 1
            if self.givenY[i] == self.Y[i] :
                classSuccess[self.Y[i]] += 1
        balancedSuccess = 0.0
        for i in range(self.numClasses) :
            if classSize[i] > 0 :
                balancedSuccess += classSuccess[i] / float(classSize[i])
        balancedSuccess /= self.numClasses
        sensitivity = 0
        ppv = 0
        if self.numClasses == 2 :
            if classSuccess[targetClass] > 0 :
                sensitivity = float(classSuccess[targetClass]) /\
                              float(classSize[targetClass])
            numTarget = Numeric.sum(Numeric.equal(self.Y, targetClass))
            if numTarget == 0 :
                ppv = 0
            else :
                ppv = float(classSuccess[targetClass]) / numTarget


        return 1 - len(self.misclassified) / float(len(self.Y)), balancedSuccess,\
               ppv, sensitivity


class Results_ (ClassificationResults) :
    """A class for holding the results of testing a classifier
    """

    plotROC = plotROC
    #computeStats = computeStats

    shortAttrList = ['info', 'log',
                     'successRate', 'balancedSuccessRate',
                     'roc','roc50',
                     'classLabels', 'confusionMatrix',
                     'ppv', 'sensitivity']
    longAttrList = ['info', 'log',
		    'Y', 'L', 'decisionFunc', 'givenY', 'givenL',
                    'classLabels',
                    'patternID', 'numClasses']

    
    def __init__(self, arg, classifier = None, **args) :

        self.rocN = 50
        if 'rocN' in args :
            self.rocN = args['rocN']

        # deal with the roc options and args :
        if 'rocTargetClass' in args :
            self.targetClass = args['rocTargetClass']
            if type(self.targetClass) == type('') :
                # the following is not optimal:
                self.targetClass = arg.labels.classDict[self.targetClass]
        else :
            self.targetClass = 1

        if 'normalization' in args :
            self.rocNormalization = args['normalization']
        else :
            self.rocNormalization = True

        if (arg.__class__.__name__ == 'Container' or
            arg.__class__ == self.__class__) :
            self.copyConstruct(arg, **args)
            return

        data = arg    
        self.Y = []
        self.L = []
        self.decisionFunc = []
        self.patternID = []
        self.givenY = []
        self.givenL = []
        
        self.successRate = 0.0
        self.info = 'dataset:\n' + data.__repr__() + \
                    'classifier:\n' + classifier.__repr__()

        if hasattr(classifier, 'labels') :
            self.classLabels = classifier.labels.classLabels
        elif data.labels.L is not None :
            self.classLabels = data.labels.classLabels
        if hasattr(self, 'classLabels') :
            self.numClasses = len(self.classLabels)

    def copyConstruct(self, other, **args) :

        if not hasattr(other, 'decisionFunc') :
            raise AttributeError, 'not a valid results object'

        if 'patterns' in args :
            p = args['patterns']
	    idDict = misc.list2dict(other.patternID, range(len(other.patternID)))
	    patterns = [idDict[pattern] for pattern in p
			if pattern in idDict]
	else :
            patterns = range(len(other.Y))

	self.patternID = [other.patternID[p] for p in patterns]
        self.L = [other.L[p] for p in patterns]
        self.Y = [other.Y[p] for p in patterns]
        self.decisionFunc = [other.decisionFunc[p] for p in patterns]
        self.givenY = [other.givenY[p] for p in patterns]
        self.givenL = [other.givenL[p] for p in patterns]        
        self.rocN = 50
        self.classLabels = copy.deepcopy(other.classLabels)
        self.numClasses = len(self.classLabels)
        self.info = other.info
        try :
            self.log = other.log
        except :
            pass
        self.computeStats()

    def __len__(self) :

        return len(self.Y)
        


    def __getattr__(self, attr) :

        if not attr.find('roc') == 0 :
            raise AttributeError, 'unknown attribute ' + attr

        if attr == 'roc' :
            rocN = None
        elif attr[-1] == '%' :        # roc1%
            rocN = float(attr[3:-1]) / 100.0
        elif float(attr[3:]) >= 1 :   # roc50
            rocN = int(float(attr[3:]))
        else :
            rocN = float(attr[3:])    # roc0.01 (equivalent to roc1%)

        rocValue = self.getROC(rocN)
        # set the value of the roc so that it does not have to be computed
        # next time it is accessed
        setattr(self, attr, rocValue)

        return rocValue
        

    def getROC(self, rocN = None) :

        rocTP, rocFP, rocValue = roc(self.Y, self.givenY, self.decisionFunc,
                                     rocN, self.targetClass, self.rocNormalization)
        return rocValue
            

    def appendPrediction(self, arg, data, pattern) :
        '''
        add the classification results and labels of a data point
        '''

        (y, f) = arg
        if f==1 and f==0 :
            return
        self.Y.append(y)
        self.decisionFunc.append(f)
	self.L.append(self.classLabels[y])
        if hasattr(data.labels, 'patternID') and data.labels.patternID is not None :
            self.patternID.append(data.labels.patternID[pattern])
        if hasattr(data.labels, 'Y') and data.labels.Y is not None :
            self.givenY.append(data.labels.Y[pattern])
            self.givenL.append(data.labels.L[pattern])


    def convert(self, *options) :

        if 'short' in options :
            attributes = self.shortAttrList
        else :
            attributes = self.longAttrList

        return convert(self, attributes)


class Results (ClassificationResults, list) :

    def __init__(self, arg = None, classifier = None, **args) :

	list.__init__(self)
        if arg.__class__ == self.__class__ or type(arg) == type([]) :
	    for r in arg :
		self.append(Results_(r, **args))
	    self.computeStats()
	elif arg is None :
	    pass
	else :
            # construct a blank object:
            self.append(Results_(arg, classifier, **args))
	    

    def __repr__(self) :

        rep = []
	for i in range(list.__len__(self)) :
            if list.__len__(self) > 1 :
                rep.append('results in fold ' + str(i))
            rep.append(self[i].__repr__())

        if list.__len__(self) > 1 :
	    rep.append(ClassificationResults.__repr__(self))

        return '\n'.join(rep)

    def __len__(self) :

	if hasattr(self, 'Y') :
	    return len(self.Y)
	else :
	    return Numeric.sum([len(res) for res in self])

    def getNumFolds(self) :
	return list.__len__(self)

    def setNumFolds(self) :
	raise AttributeError

    numFolds = property(getNumFolds, setNumFolds, None, 'number of folds')
        
    def __getattr__(self, attr) :

	if attr.find('roc') == 0 :
	    return Numeric.average([getattr(results, attr)
			     for results in self])
	elif hasattr(self[0], attr) :
	    if list.__len__(self) == 1 :
		return getattr(self[0], attr)
	    else :
		return [getattr(results, attr) for results in self]
	else :
            raise AttributeError, 'unknown attribute ' + attr

    def computeStats(self) :

        for res in self :
            res.computeStats()
        
        if list.__len__(self) > 1 :
            self.poolData()
            ClassificationResults.computeStats(self)


    def plotROC(self, fileName=None, foldNum = None, **args) :

        if foldNum is None : foldNum = 0
        if foldNum > list.__len__(self) :
            raise ValueError, 'foldNum too large'
        
        self[foldNum].plotROC(fileName, **args)

    def poolData(self) :

        self.Y = []
        self.givenY = []
        self.L = []
        self.givenL = []
        self.decisionFunc = []
        self.patternID = []
	self.numClasses = self[0].numClasses
        self.classLabels = self[0].classLabels
	if not hasattr(self, 'rocN') : self.rocN = self[0].rocN
	if not hasattr(self, 'targetClass') :
            self.targetClass = self[0].targetClass
	if not hasattr(self, 'rocNormalization') : 
	    self.rocNormalization = self[0].rocNormalization
        for res in self :
            self.Y.extend(res.Y)
            self.givenY.extend(res.givenY)
            self.L.extend(res.L)
            self.givenL.extend(res.givenL)
            self.decisionFunc.extend(res.decisionFunc)
            self.patternID.extend(res.patternID)


    def appendPrediction(self, arg, data, pattern) :

        self[-1].appendPrediction(arg, data, pattern)

    def append(self, other) :

	if other.__class__.__name__ == 'Results_' :
	    list.append(self, other)
	elif other.__class__ == self.__class__ :
	    for res in other :
		list.append(self, res)
	else :
	    raise ValueError, 'wrong type of object: we are not the same type'


    def convert(self, *options) :

        return [results.convert(*options) for results in self]
    
    def save(self, fileName, *options) :
        """
	save Results to a file
        only attributes given in the attribute list are saved

        OPTIONS::

          'long' - use long attribute list
	  'short' - use short attribute list
	  using the short attribute list won't allow you to reconstruct
	  the Results object afterwards, only the statistics that characterize
          the results.
        """

        resultsList = self.convert(*options)

	myio.save(resultsList, fileName)

    def toFile(self, fileName, delim = '\t') :
	"""
	save results to a (tab) delimited file

	format is:
	patternID, decision function, predicted class, given class, fold

	:Parameters:
	  - `fileName` - file name to which to save the results
	  - `delim` - delimiter (default: tab)
	"""
	
	outfile = open(fileName, 'w')
	for fold in range(self.numFolds) :
	    results = self[fold]
	    for i in range(len(results)) :
		outfile.write(
		    delim.join([results.patternID[i],
				str(results.decisionFunc[i]),
				results.L[i],
				results.givenL[i],
				str(fold + 1)]) + '\n')
			  

    def get(self, attribute, fold = None) :

        if fold is None :
            if self.numFolds == 1 :
                return getattr(self[0], attribute)
            else :
                return getattr(self, attribute)
        else :
            return getattr(self[fold], attribute)


    def getPredictedLabels(self, fold = None) :

        return self.get('L', fold)

    def getPredictedClass(self, fold = None) :

        return self.get('Y', fold)        

    def getGivenClass(self, fold = None) :

        return self.get('givenY', fold)        
        
    def getGivenLabels(self, fold = None) :

        return self.get('givenL', fold)
        
    def getDecisionFunction(self, fold = None) :

        return self.get('decisionFunc', fold)

    def getPatternID(self, fold = None) :

        return self.get('patternID', fold)

    def getLog(self, fold = None) :

	if fold is None :
	    return [self.get('log', fold_) for fold_ in range(self.numFolds)]
	else :
	    return self.get('log', fold)        

    def getInfo(self, fold = None) :

	if fold is None :
	    return [self.get('info', fold_) for fold_ in range(self.numFolds)]
	else :
	    return self.get('info', fold)

    def getROC(self, fold = None) :

        return self.get('roc', fold)

    def getROCn(self, rocN = None, fold = None) :

        if rocN is None : rocN = self.rocN
        return self.get('roc' + str(rocN), fold)

    def getSuccessRate(self, fold = None) :

        return self.get('successRate', fold)

    def getBalancedSuccessRate(self, fold = None) :

        return self.get('balancedSuccessRate', fold)

    def getConfusionMatrix(self, fold = None) :

        return self.get('confusionMatrix', fold)

    def getPPV(self, fold = None) :

        return self.get('ppv', fold)

    def getSensitivity(self, fold = None) :

        return self.get('sensitivity', fold)

    def getClassLabels(self) :

        return self.classLabels
    
def convert (object, attributes) :

    obj = misc.Container()
    obj.addAttributes(object, attributes)

    return obj


def saveResultObjects (objects, fileName, *options) :
    """
    save a list or dictionary of Results objects
    it is o.k. if the list or dictionary is itself a list or dictionary of 
    OPTIONS:
    long - save the long attribute list
    """
    
    if type(objects) == type([]) :
	if type(objects[0]) == type([]) :
	    obj = [ [o.convert(*options) for o in resultsList]
		    for resultsList in objects]
	elif type(objects[0]) == type({}) :
	    obj = []
	    for resultsList in objects :
		object = {}
	else :
	    obj = [o.convert(*options) for o in objects]

    elif type(objects) == type({}) :
	obj = {}
	for rkey in objects :
	    if type(objects[rkey]) == type({}) :
		obj[rkey] = {}
		for key in objects[rkey] :
		    obj[rkey][key] = object[rkey][key].convert(*options)
	    elif type(objects[rkey]) == type([]) :
		obj[rkey] = [ object.convert(*options) for object in objects[rkey] ]
	    else :
		obj[rkey] = objects[rkey].convert(*options)
    else :
        raise ValueError, 'expected either a list or dictionary'

    myio.save(obj, fileName)



class ResultsList (list) :

    def __init__(self, resList = None) :
	
	self.rocN = 50
	if resList is None : return
	for results in resList :
	    if type(results) == type([]) :
		self.append(Results(results))
	    else :
		self.append(results)
	self.computeStats()

    def __repr__(self) :

	rep = []

	rep.append('number of Results objects: %d' % len(self))
        rep.append('success rate: %f %f' % (self.successRate, MLab.std(self.successRates)) )
        rep.append('balanced success rate: %f %f' % 
		   (self.balancedSuccessRate, MLab.std(self.balancedSuccessRates)) )

        rep.append('area under ROC curve: %f %f' % (self.roc, MLab.std(self.rocs)) )
	rep.append('area under ROC %d curve: %f %f' % \
		   (self.rocN, Numeric.average(self.rocNs), MLab.std(self.rocNs)))

	return '\n'.join(rep)

    def save(self, fileName, *options) :

        if 'short' in options :
            attributes = self[0][0].shortAttrList
        else :
            attributes = self[0][0].longAttrList

	resultsList = [[convert(results, attributes) for results in res]
		       for res in self]

	myio.save(resultsList, fileName)

    def computeStats(self) :

	self.balancedSuccessRates = [res.balancedSuccessRate for res in self]
	self.balancedSuccessRate = Numeric.average(self.balancedSuccessRates)
	self.successRates = [res.successRate for res in self]
	self.successRate = Numeric.average(self.successRates)
	self.rocs = [res.roc for res in self]
	self.roc = Numeric.average(self.rocs)
	self.rocNs = [res.getROCn(self.rocN) for res in self]


class RegressionResults (object) :

    def __init__(self, data, classifier, **args) :

        self.Y = []
        self.givenY = []
        self.patternID = []

    def appendPrediction(self, y, data, pattern) :

        self.Y.append(y)
        if hasattr(data.labels, 'patternID') and data.labels.patternID is not None :
            self.patternID.append(data.labels.patternID[pattern])
        if hasattr(data.labels, 'Y') and data.labels.Y is not None :
            self.givenY.append(data.labels.Y[pattern])

    def extend(self, other) :

        self.Y.extend(other.Y)
        self.givenY.extend(other.givenY)
        self.patternID.extend(other.patternID)
        
    def computeStats(self) :

        pass


def loadResults(fileName, isNewFormat = True) :
    """
    isNewFormat -- whether the Results were saved under version 0.6.1 or newer
    """

    res = myio.load(fileName)
    if not isNewFormat :
        return Results([res])
    if type(res) == type({}) :
        results = {}
        for key in res :
            results[key] = Results(res[key])
        return results
    if type(res[0]) == type([]) :
	return ResultsList(res)
    else :
	return Results(res)

def loadResults2(fileName) :
    """
    load a list of list of Results objects or a dictionary of a list of Results objects
    """
    
    res = myio.load(fileName)
    if type(res) == type({}) :
        results = {}
        for key in res :
            results[key] = [ Results(object) for object in res[key] ]
	return results

    elif type(res) == type([]) :
	return [ [Results(object) for object in listOfResults] for listOfResults in res ]


