
import os

from ext import ckernel
from ker import Kernel

__docformat__ = "restructuredtext en"

"""functionality for dealing with C++ kernel objects"""


class Linear (Kernel, ckernel.Linear) :
    '''A Linear kernel (dot product)
    Construction:
    k = Linear()
    Attributes:
    type - kernel type ("LINEAR")
    Methods:
    eval(x,y) - evaluate the dot product between vectors x and y
    '''

    type = 'kernel'
    
    def __init__(self, arg = None) :
        
        ckernel.Linear.__init__(self)

    def constructionParams(self) :

        return ''
    
    def eval (self, datai, i, j, dataj = None) :

        if dataj is None :
            return datai.dotProduct(i, j)
        else :
            return datai.dotProduct(i, j, dataj)

class Cosine (Kernel, ckernel.Cosine) :
    '''A Cosine kernel (dot product)
    Construction:
    k = Cosine()
    Attributes:
    type - kernel type ("LINEAR")
    Methods:
    eval(x,y) - evaluate the dot product between vectors x and y
    '''
    type = 'kernel'
    
    def __init__(self, arg = None) :
        
        ckernel.Cosine.__init__(self)
    
    def constructionParams(self) :

        return ''
    
    def eval (self, datai, i, j, dataj = None) :

        if dataj is None : dataj = datai
        return ckernel.Cosine.eval(self,
                                   datai.castToBase(), i, j, dataj.castToBase())


class Polynomial (Kernel, ckernel.Polynomial) :
    '''A Polynomial kernel
    K(x,y) = (x \dot y + additiveConst) ** degree
    Construction:
    k = Polynomial(degree, additiveConst)
    Attributes:
    additiveConst, degree - kernel parameters
    Methods:
    eval(x,y) - evaluate the kernel between vectors x and y
    '''
    def __init__(self, arg = 2, **args):

        additiveConst = 1.0
        if arg.__class__.__name__ == self.__class__.__name__ :
            degree = arg.degree
            additiveConst = arg.additiveConst
        elif type(arg) == type(2) :
            degree = int(arg)
        else :
            raise ValueError, 'wrong type of argument for Polynomial'

        if 'degree' in args :
            degree = int(args['degree'])
        if 'additiveConst' in args :
            additiveConst = float(args['additiveConst'])

        ckernel.Polynomial.__init__(self, degree, additiveConst)

    def __repr__(self) :

        rep = '<' + self.__class__.__name__ + ' instance>\n'
        rep += 'degree : ' + str(self.degree) + '\n'
        rep += 'affine coefficient : ' + str(self.additiveConst)
        
        return rep

    def constructionParams(self) :

        return 'degree = ' + str(self.degree) + ',' + \
               'additiveConst = ' + str(self.additiveConst)

    def eval (self, datai, i, j, dataj = None) :

        if dataj is None : dataj = datai
        return ckernel.Polynomial.eval(self,
                                       datai.castToBase(), i, j, dataj.castToBase())

class Gaussian (Kernel, ckernel.Gaussian) :

    '''A Gaussian (RBF) kernel
    K(x,y) = exp( - gamma * ||x - y||**2
    Construction:
    k = Gaussian(gamma)
    Attributes:
    gamma - kernel width parameter
    Methods:
    eval(x,y) - evaluate the kernel between vectors x and y
    '''

    def __init__(self, arg = 1.0, **args) :

        gamma = 1.0
        if arg.__class__.__name__ == self.__class__.__name__ :
            gamma = arg.gamma
        elif type(arg) == type(1.0) or type(arg) == type(1) :
            gamma = float(arg)
        else :
            raise ValueError, 'wrong type of argument for Gaussian'

        if 'gamma' in args :
            gamma = float(args['gamma'])
        ckernel.Gaussian.__init__(self, gamma)

    def __repr__(self) :
        
        rep = '<' + self.__class__.__name__ + ' instance>\n'
        rep += 'gamma : ' + str(self.gamma)
        
        return rep

    def constructionParams(self) :

        return 'gamma = ' + str(self.gamma)

    def eval (self, datai, i, j, dataj = None) :

        if dataj is None : dataj = datai
        return ckernel.Gaussian.eval(self,
                                     datai.castToBase(), i, j, dataj.castToBase())


def attachKernel(data, kernel = 'linear', **args) :

    print 'attach kernel for a wrapper dataset'
    if type(kernel) == type('') :
        kernel = kernel.lower()
        if kernel == 'linear' or kernel == 'lin' :
            k = Linear(**args)
        elif kernel == 'cosine' :
            k = Cosine(**args)
        elif kernel == 'polynomial' or kernel == 'poly' :
            k = Polynomial(**args)
        elif kernel == 'rbf' or kernel == 'gaussian' :
            k = Gaussian(**args)
        else :
            raise ValueError, 'unrecognized type of kernel'
    elif hasattr(kernel, 'type') and kernel.type == 'dataset' :
        print 'copying kernel'
        other = kernel
        k = other._kernel.__class__(other._kernel)
    elif hasattr(kernel, 'type') and kernel.type == 'kernel' :
        k = eval(kernel.__class__.__name__ + '(kernel)')

    # destroy the data object's kernel if it has one:
    if hasattr(data, '_kernel') :
        # make sure that the C++ object is destroyed:
        data._kernel.thisown = 1
        #    del data._kernel
    data._kernel = k
    data._kernel.thisown = 0
    data.setKernel(k.castToBase())


def kernel2file(data, fileName, **args) :

    """compute a kernel matrix and save it to a file in tab delimited format

    :Parameters:
      - `data` - a dataset
      - `fileName` - file name to save the kernel

    :Keywords:
      - `format` - whether to save the matrix in pyml or gist formats
    """

    format = 'pyml'
    if 'format' in args :
        format = args['format']
    import tempfile
    tmpfile = tempfile.mktemp()
    ckernel.kernel2file(data.castToBase(), tmpfile)
    tmp = open(tmpfile)
    file = open(fileName, 'w')
    if format == 'gist' :
        file.write(fileName + '\t')
        file.write('\t'.join(data.labels.patternID) + '\n')
    i = 0
    for line in tmp :
        if data.labels.patternID is not None :
            file.write(data.labels.patternID[i])
        i += 1
        file.write(line)
    os.remove(tmpfile)

def averageEntry(fileName, ignoreDiagonal = True, delim = None) :

    s = 0
    numEntries = 0
    file = open(fileName)
    i = 0
    for line in file :
        tokens = line.split(delim)
        for token in tokens :
            try :
                val = float(token)
                s += val
                numEntries += 1
            except :
                pass
        i += 1
    return s / numEntries
