
import Numeric
import MLab
import os

import datafunc
import misc
from baseClassifiers import Classifier
from composite import CompositeClassifier
import assess

"""classes for multi-class classification"""

__docformat__ = "restructuredtext en"

class OneAgainstOne (CompositeClassifier) :

    '''One-against-one Multi-class classification
    using a two class classifier.
    
    For a k class problem k(k-1) binary classes are trained for all
    pairs of classes; an instance is classified to the class that
    receives the highest number of votes; an instance is constructed
    using a classifier that is used as a template for constructing
    the actual classifiers.
    '''

    def train(self, data, **args) :
        '''train k(k-1)/2 classifiers'''

        Classifier.train(self, data, **args)
        numClasses = self.labels.numClasses

        if numClasses <= 2:
            raise ValueError, 'Not a multi class problem'

        self.classifiers = misc.matrix((numClasses, numClasses))
        for i in range(numClasses - 1) :
            for j in range(i+1, numClasses) :
                self.classifiers[i][j] = self.classifier.__class__(self.classifier)
                dataij=data.__class__(data, deepcopy = self.classifier.deepcopy,
                                      classID = [i,j])
                self.classifiers[i][j].train(dataij)
        self.log.trainingTime = self.getTrainingTime()                

    def classify(self, data, p):
        '''Suppose that x is classified to class c, then the margin is
        defined as the minimum margin found against the k-1 other classes
        '''
        
        numClasses = self.labels.numClasses
        r = Numeric.zeros((numClasses, numClasses),Numeric.Float)
        vote = Numeric.zeros(numClasses)
        for i in range(numClasses - 1) :
            for j in range(i+1, numClasses) :
                r[i][j] = self.classifiers[i][j].decisionFunc(data, p)
                # afterwards I take the minimum or r, so assign:
                r[j][i] = r[i][j]  
                if r[i][j] > 0 :
                    vote[j] += 1
                else:
                    vote[i] += 1
                    
        maxvote = Numeric.argmax(vote)
        return maxvote, MLab.min(MLab.absolute(r[maxvote]))

    def preproject(self, data) :

        for i in range(self.labels.numClasses-1):
            for j in range(i+1, self.labels.numClasses):
                self.classifiers[i][j].preproject(data)
        
    test = assess.test

    
class OneAgainstRest (CompositeClassifier) :

    '''A one-against-the-rest multi-class classifier'''
    
    def train(self, data, **args) :
        '''train k classifiers'''

        Classifier.train(self, data, **args)
        if 'labels' in args :
            raise ValueError, 'optional args not implemented for this class'

        numClasses = self.labels.numClasses
        if numClasses <= 2:
            raise ValueError, 'Not a multi class problem'

        self.classifiers = [self.classifier.__class__(self.classifier)
                            for i in range(numClasses)]

        for i in range(numClasses) :
            datai = data.__class__(data, deepcopy = self.classifier.deepcopy)
            datai.attachLabels(datafunc.oneAgainstRest(data.labels,
                                                       data.labels.classLabels[i]))
            self.classifiers[i].train(datai)
            #data.attachLabels(self.labels)
        self.log.trainingTime = self.getTrainingTime()
        
    def classify(self, data, i):

        r = Numeric.zeros(self.labels.numClasses, Numeric.Float)
        for j in range(self.labels.numClasses) :
            r[j] = self.classifiers[j].decisionFunc(data, i)

        return MLab.argmax(r), MLab.max(r)

    def preproject(self, data) :

        for i in range(self.labels.numClasses) :
            self.classifiers[i].preproject(data)

    test = assess.test


def allOneAgainstRest(classifier, data, resultsFile, numFolds = 5, minClassSize = 8) :

    import myio
    labels = data.labels
    
    results = {}
    if os.path.exists(resultsFile) :
	results = assess.loadResults(resultsFile)
    
    for label in labels.classLabels :
        if label in results : continue
        if (minClassSize is not None and
            labels.classSize[labels.classDict[label]] < minClassSize) : continue
        
        myio.log('class: ' + label + '\n')        
        data = datafunc.oneAgainstRest(data, label)
	try :
	    results[label] = classifier.stratifiedCV(data, numFolds)
	except :
	    results[label] = None
        assess.saveResultObjects(results, resultsFile)
        data.attachLabels(labels)
    
