
import Numeric
import MLab
import math
import misc
import random
import datafunc

def pca(X, numcomp = None) :
    '''returns the matrix X as represented in the numcomp leading principal
    components
    if numcomp is None, all principal components are returned'''
    
    d = Numeric.shape(X)[1]
    if numcomp is None :
        numcomp = d
        
    [u,s,v] = MLab.svd(X)
    v = Numeric.transpose(v)
    v = v[:,:numcomp]
    print Numeric.shape(X)
    return Numeric.matrixmultiply(X, v)


def centerColumns(X) :
    '''returns X - mean(X), where the mean is taken over the columns of X'''
    
    m = MLab.mean(X)

    n = Numeric.shape(X)[0]
    
    return X - Numeric.resize(m, (n,len(m)))


def centerRows(X) :

    return Numeric.transpose(centerColumns(Numeric.transpose(X)))


def standardizeColumns(X) :
    '''returns (X - mean(X)) / std(X) '''

    m = MLab.mean(X)
    std = MLab.std(X)

    n = Numeric.shape(X)[0]

    return (X - Numeric.resize(m, (n,len(m))))/Numeric.resize(std, (n,len(std)))
    
def standardizeRows(X) :
    '''returns (X - mean(X)) / std(X) '''

    return Numeric.transpose(standardizeColumns(Numeric.transpose(X)))
    

def maxvar(X, numVariables) :
    '''returns the numVariables variables with the highest variance'''
    
    s = MLab.std(X)
    I = Numeric.argsort(s)
    
    Xout = Numeric.take(X, I[-numVariables:], 1)

    return Xout


def dmat(X) :
    '''returns the Euclidean distance-squared matrix'''

    K = Numeric.matrixmultiply (X, Numeric.transpose (X))
    n = Numeric.shape(K)[0]
    D = Numeric.zeros((n,n), Numeric.Float)

    for i in range(1, n-1) :
        for j in range(i+1, n) :
            D[i,j] = K[i,i] - 2 * K[i,j] + K[j,j]
            D[j,i] = D[i,j]
    
    return D


def norm2(x) :
    '''return the 2-norm of a vector given as a list or Numeric array'''
    
    x = Numeric.asarray(x)
    
    return math.sqrt(Numeric.sum(x*x))


def normalizeNorm(X) :
    '''normalize each row of X to unit vectors'''

    (numRows, numCols) = Numeric.shape(X)
    Xnorm = Numeric.zeros((numRows, numCols), Numeric.Float)
    
    for i in range(numRows) :
        Xnorm[i] = X[i] / norm2(X[i])

    return Xnorm
    
class Correlator (object) :

    def __init__(self, data) :

        if type(data) == type('') :
            print 'file name:', data            
            data = datafunc.DataSet(data, idColumn = 0, headerRow = True)

        self.data = data
        self.idDict = misc.list2dict(data.labels.patternID,
                                     range(len(data)))

        print Numeric.shape(data.X)
        self.mean = MLab.mean(data.X, 1)
        self.std = std(data.X, 1)
        eps = 1e-5
        I = Numeric.nonzero(Numeric.less(self.std, eps))
        print 'num zeros:',len(I)
        Numeric.put(self.std, I, 1)
        
        self.numCorrelations = 10000
        correlations = Numeric.zeros(self.numCorrelations, Numeric.Float)
        
        for i in range(self.numCorrelations) :
            i1 = random.randrange(0, len(data))
            i2 = random.randrange(0, len(data))
            correlations[i] = self._corrcoef(i1, i2)
        self.meanCorrelation = MLab.mean(correlations)
        self.numCorrelations = 1000        

    def corrcoef(self, id1, id2) :

        if id1 == id2 : return 1.0
        if type(id1) == type(1) :
            return self._corrcoef(id1, id2)
        if id1 not in self.idDict and id2 not in self.idDict :
            return self.meanCorrelation
        if id1 in self.idDict and id2 in self.idDict :
            return self._corrcoef(self.idDict[id1], self.idDict[id2])
        else :
            # we want to assume that id1 not in data:
            if id2 not in self.idDict :
                id1,id2 = id2,id1
            i2 = self.idDict[id2]
            correlations = Numeric.zeros(self.numCorrelations, Numeric.Float)
            for i in range(self.numCorrelations) :
                i1 = random.randrange(0, len(self.data))
                correlations[i] = self._corrcoef(i1, i2)
            return MLab.mean(correlations)

    def _corrcoef(self, i1, i2) :
        
        return Numeric.dot(self.data.X[i1] - self.mean[i1],
                           self.data.X[i2] - self.mean[i2]) / \
                           (len(self.data.X[i1]) * self.std[i1] * self.std[i2])
    
    
def corrcoef2(X) :
    '''compute the correlation between the rows of the matrix X
    more space efficient than MLab version'''
    
    (n,d) = Numeric.shape(X)

    m = MLab.mean(X, 1)
    std = MLab.std(X, 1)

    K = Numeric.ones((n,n), Numeric.Float)

    for i in range(0, n - 1) :
        for j in range(i + 1, n) :
            K[i][j] = Numeric.dot(X[i] - m[i], X[j] - m[i]) / (d * std[i] * std[j])
            K[j][i] = K[i][j]

    return K

def std(m,axis=0):
    """std(m,axis=0) returns the standard deviation along the given 
    dimension of m.  The result is unbiased with division by N-1.
    If m is of integer type returns a floating point answer.
    """
    x = Numeric.asarray(m)
    n = float(x.shape[axis])
    mx = Numeric.asarray(MLab.mean(x,axis))
    if axis < 0:
        axis = len(x.shape) + axis
    mx.shape = mx.shape[:axis] + (1,) + mx.shape[axis:]
    x = x - mx
    return Numeric.sqrt(Numeric.add.reduce(x*x,axis)/(n))

def corrcoef(X) :

    (n,d) = Numeric.shape(X)

    Xn = standardizeRows(X)

    return Numeric.dot(Xn, Numeric.transpose(Xn)) / (d - 1)

def corrcoefij(X, i, j) :

    (n,d) = Numeric.shape(X)

    m = MLab.mean(X, 1)
    std = MLab.std(X, 1)


    return Numeric.dot(X[i] - m[i], X[j] - m[i]) / (d * std[i] * std[j])


    


class Rescale (object) :
    """
    class for performing feature normalization
    
    this rescaling is composed of two operations:
    
      1.  ``centering`` -- subtract from a feature its mean value;
          this is referred to as 'translation'; the translation attribute
          gives the value with which to translate each feature
      2.  ``scaling`` -- divide a feature by a scale, e.g. its standard deviation;
          the 'scale' attribute gives the value with which to scale each feature

    the 'train' method of the class computes the translation and scaling
    factors, and performs normalization of the training data
    the 'test' method uses values computed on the training data to normalize
    the test data.
    
    **caveat:**
    Beware of performing training multiple times on the same dataset:
    if a dataset has already been standardized, re-standardization
    will recompute mean and standard deviation, which will be approximately
    0 and 1 for each feature; subsequent application on test data will
    have no effect. Because of this an exception is raised if the user
    attempts to re-train an already trained Rescale object.
    """


    def __init__(self, arg = None) :

        self.translate = True
        self.rescale = True
        self.translation = None
        self.scale = None

    def train(self, data, *options, **args) :

        if self.translation is not None or self.scale is not None :
            raise ValueError, 'object already trained'
        if self.translate :
            self.translation = data.mean()
        if self.rescale :
            self.scale = Numeric.array(data.std())
            #self.features = data.featureID[:]
            # need to avoid division by 0, so
            # scales that are equal to 0 are replaced with a value of 1
            eps = 1e-5
            I = Numeric.nonzero(Numeric.less(self.scale, eps))
            Numeric.put(self.scale, I, 1)
            # checking for nan:
            for i in range(len(self.scale)) :
                if self.scale[i] == 0 and self.scale[i] == 1 :
                    self.scale[i] = 1
        self.preproc(data)

    def preproc(self, data) :

        if self.translate :
            data.translate(self.translation)
        if self.rescale :
            data.scale(1.0 / self.scale)

    def test(self, data, *options, **args) :

        print 'preprocessing test data'
        self.preproc(data)
    
