import os
import random
import Numeric

import misc
from baseClassifiers import Classifier
from ext.libsvm import C_SVC, NU_SVC, ONE_CLASS, EPSILON_SVR, NU_SVR
from ext.libsvm import LINEAR, POLY, RBF, SIGMOID, LINEARRIDGE
from ext import libsvm
import libsvmWrap
from ext import csvmodel
import arrayWrap
import assess
import datafunc
import ker,kernelWrap

from ext import csmo
from ext import cgist


"""various flavors of SVMs and training algorithms"""

__docformat__ = "restructuredtext en"

class SVM (Classifier) :
    """
    An SVM classifier class.

    SVM is trained using either libsvm, or using a PyML SMO implementation
    based on libsvm
    """

    svm_type = C_SVC
    attributes = {'C' : 10,
                  'Cmode': 'classProb',
                  'optimizer' : 'libsvm',
                  'cacheSize' : 256,
                  'nu' : 0.1,
                  'eps' : 0.01}
    
    def __init__(self, arg = None, **args):

        """
        :Parameters:
          - `arg` - another SVM object or a kernel object; if no argument is given
            the kernel function of the training dataset is used
        
        :Keywords:
          - `C` - the svm C parameter
          - `Cmode` - the way the C parameter is used; values: 'equal', 'classProb',
            'fromData'.
            in 'equal' mode C is set to be the same for both classes
            in 'classProb' mode C is made proportional to the class prior
            probability (fraction of each class in the dataset).  This is useful
            for datasets with an unbalanced class distribution
            in 'fromData' the value of C for each pattern is taken from the
            'C' attribute of the training data.
          - `optimizer` - which optimizer to use.  values: 'libsvm' -- run libsvm
            'mysmo' - use the PyML native optmizer (based on libsvm)
            'gist' - use a gist-like optimizer.
          - `cacheSize` - size of the kernel cache (in MB).
        """

        Classifier.__init__(self, arg, **args)

        self.kernel = None
        if arg.__class__ == self.__class__ :
            if hasattr(arg, 'ridge') :
                self.ridge = other.ridge[:]
            if arg.kernel is not None :
                self.kernel = arg.kernel.__class__(arg.kernel)
        elif hasattr(arg, 'type') and arg.type == 'kernel' :
            self.kernel = arg.__class__(arg)
            
    def __repr__(self) :

        rep = ['<' + self.__class__.__name__ + ' instance>']
        if hasattr(self, 'C') :
            rep.append('C : %f' % self.C)
            rep.append('Cmode: %s' % self.Cmode)
        if hasattr(self, 'kernel') and self.kernel is not None :
            rep.append(str(self.kernel))
        if hasattr(self, 'model') :
            if hasattr(self, 'model') :
                rep.append(str(self.model))
        
        return '\n'.join(rep)

    def save(self, fileName) :

        """
        save an SVM model to a file.
        use the loadSVM method to then load the saved model
        be sure the call the SVM train function as:
        train(data, saveSpace=False)
	
	:Parameters:
	  - `fileName` - a file name or file handle
        """

        self.model.save(fileName)
            
    def train(self, data, **args) :

        """
        train an SVM

        :Keywords:
          - `saveSpace` -- whether to save memory when constructing an SVM model
            [default: True]
            you need to set this keyword to False if you want to save the
            resulting model
        """
    
        Classifier.train(self, data, **args)
        if self.kernel is not None :
            if (self.kernel.__class__.__name__ == 'LinearRidge') :
                data.attachKernel('linear')
            else :
                data.attachKernel(self.kernel)
            
        if self.optimizer == 'libsvm' :
            alpha,b,svID = self.trainLibsvm(data, **args)
        elif self.optimizer == 'gist' :
            alpha,b,svID = self.trainGist(data, **args)
        elif self.optimizer == 'gradient' :
            alpha,b,svID = self.trainGradient(data, **args)            
        else :
            alpha,b,svID = self.trainMySMO(data, **args)

        if (data.kernel.__class__.__name__.find('Linear') == 0
            and data.isVector) :
            if data.isWrapper :
                self.model = LinearSVModelWrap(data, svID, alpha, b, **args)
            else :
                self.model = LinearSVModel(data, svID, alpha, b, **args)
        else :
            if data.isWrapper :
                self.model = SVModelWrap(data, svID, alpha, b, **args)
            else :
                self.model = SVModel(data, svID, alpha, b, **args)

        self.trained = True
        self.log.numSV = len(alpha)
        self.log.trainingTime = self.getTrainingTime()

        
    def trainLibsvm(self, data, **args) :
            
        # setting C for the positive and negative classes
        if (self.svm_type == ONE_CLASS or
            self.svm_type == EPSILON_SVR or
            self.svm_type == NU_SVR) :
            Cpos = 0
            Cneg = 0
        else :
            if data.labels.numClasses != 2 :
                raise ValueError, 'svm is a two class classifier'
            if self.Cmode == "classProb":
                Cpos = self.C * (float(data.labels.classSize[0]) / float(len(data)))
                Cneg = self.C * (float(data.labels.classSize[1]) / float(len(data)))
            else:
                Cpos = Cneg = self.C

        print 'Cpos, Cneg: ', Cpos,Cneg

        # prepare data for the libsvm wrapper :
        # set kernel:
        if hasattr(self, 'kernel') and self.kernel is not None :
            kernel = self.kernel
        else :
            kernel = data.kernel
        kernelType = kernel.__class__.__name__
        #param = libsvm.svm_parameter(kernel_type = LINEAR)
        param = libsvmWrap.svmParameter(kernel_type = LINEAR,
                                        svm_type = self.svm_type)
        param.C = self.C
        param.nu = self.nu
        param.shrinking = 1
        cridge = None
        if kernelType == "Polynomial" :
            # (gamma x' y + coef0)^degree
            param.kernel_type = POLY
            param.degree = kernel.degree
            param.coef0 = kernel.additiveConst
            param.gamma = 1
        elif kernelType == "Gaussian":
            # exp(-gamma * |x - y|^2)
            param.kernel_type = RBF
            param.gamma = kernel.gamma
        elif kernelType == "Cosine" :
	    # i'm using the sigmoid kernel as the cosine kernel
            param.kernel_type = SIGMOID
        elif kernelType == "LinearRidge" :
            print 'using linear ridge kernel'
            param.kernel_type = LINEARRIDGE
            if not hasattr(self, 'ridge') :
                RidgeList = [1.0/Cneg, 1.0/Cpos]
                #Cpos = Cneg = 1e10
                self.ridge = [RidgeList[data.labels.Y[i]]
                              for i in range(len(data))]
            cridge = arrayWrap.doubleArray(self.ridge)
            param.ridge = cridge.cast()
            
        s=libsvm.DecisionFunction()
        if data.__class__.__name__ == 'SparseCDataSet' :
            data.convert2libsvm()
            prob = data.prob
            libsvm.svm_train_one4(prob, param.this, Cpos, Cneg, s.this)
            data.libsvmDestroy()
        else :
            prob=libsvmWrap.svmProblem (data)
            libsvm.svm_train_one4(prob.this, param.this, Cpos, Cneg, s.this) 
            del prob

        if cridge is not None :
            del cridge

        b = - s.rho

        numSV = s.numSV
        alphaArray = libsvm.doubleArray_frompointer(s.alpha)
        svArray = libsvm.intArray_frompointer(s.svID)
        svID = [svArray[i] for i in range(numSV)]
        alpha = [alphaArray[i] for i in range(numSV)]

        return alpha, b, svID

    def getC(self, data) :

        if self.Cmode == "fromData" :
            C = data.C
        elif self.Cmode == "classProb":
            Cpos = self.C * (float(data.labels.classSize[0]) / float(len(data)))
            Cneg = self.C * (float(data.labels.classSize[1]) / float(len(data)))
            c = [Cneg, Cpos]
            C = [c[data.labels.Y[i]] for i in range(len(data))]
        else:
            C = [self.C for i in range(len(data))]

        return C

    def trainGist(self, data, **args) :

        if data.labels.numClasses != 2 :
            raise ValueError, 'svm is a two class classifier'

        alpha, b = runGist(self, data)

        svID = [i for i in range(len(alpha))
                if alpha[i] > 0]
        alpha = [alpha[i] * (data.labels.Y[i] * 2 - 1) for i in range(len(alpha)) if alpha[i] > 0]

        return alpha, b, svID

    def trainGradient(self, data, **args) :

        if data.labels.numClasses != 2 :
            raise ValueError, 'svm is a two class classifier'

        alpha, b = runGradientDescent(self, data)

        svID = [i for i in range(len(alpha))
                if alpha[i] > 0]
        alpha = [alpha[i] * (data.labels.Y[i] * 2 - 1) for i in range(len(alpha)) if alpha[i] > 0]

        return alpha, b, svID
        
    
    def trainMySMO(self, data, **args) :

        if data.labels.numClasses != 2 :
            raise ValueError, 'svm is a two class classifier'

        alpha, b = runMySMO(self, data)

        svID = [i for i in range(len(alpha))
                if alpha[i] > 0]
        alpha = [alpha[i] * (data.labels.Y[i] * 2 - 1) for i in range(len(alpha)) if alpha[i] > 0]

        b = - b

        return alpha, b, svID

    def decisionFunc(self, data, i) :

        if data.isWrapper :
            return self.model.decisionFunc(data, i)

        if self.projectionRequired :
            x = data.X[i]
            xproj = self.project(x)
        else :
            xproj = data.X[i]
        return self.model.decisionFunc(xproj)


    def classify(self, data, i) :

        margin = self.decisionFunc(data, i)
        if margin > 0 :
            return (1,margin)
        else:
            return (0,margin)


def loadSVM(fileName, **args) :

    """
    returns a trained SVM object constructed from a saved SVM model.

    The saved SVM model stores the support vectors in sparse
    vector format.  When creating the model it then represents the
    support vectors in some dataset container.  The type of the
    container needs to agree with the type of dataset of your test
    data.  By default the support vectors are represented using the
    SparseCDataSet container. You can set this using the 'datasetClass'
    keyword argument e.g. datasetClass = datafunc.DataSet
    """
    
    if 'datasetClass' in args :
        datasetClass = args['datasetClass']
    else :
        datasetClass = datafunc.SparseCDataSet

    data = None
    infile = open(fileName)
    for line in infile :
        if line[0] != '#' : break
	#if line.find('classLabels=') > 0 :
	#    classLabels = line.split('=')[1].split()
	#if line.find('numClasses') > 0 :
	#    numClasses = int(line.split('=')[1])
        if line.find('b=') > 0 :
            b = float(line[3:])
        if line.find('alpha=') > 0 :
            tokens = line.split('=')[1].split()
            alpha = [float(token) for token in tokens]
        if line.find('k=') > 0 :
            exec 'kernel = ' + line.split('k=')[1]
    infile.close()
    
    data = datasetClass(fileName, **args)
    data.attachKernel(kernel)
    # defining a nonelinear kernel is not enough, needed to do the attach kernel for the data.
    s = SVM(kernel)
    s.labels = misc.Container()
    s.labels.addAttributes(data.labels, ['numClasses', 'classLabels'])
    s.featureID = data.featureID[:]
    if datasetClass == datafunc.SparseCDataSet :
	print 'loading using wrapper data'
        if (kernel.__class__.__name__ == 'Linear' or
            kernel.__class__.__name__ == 'Cosine') :
            s.model = LinearSVModelWrap(data, range(len(data)), alpha, b, **args)
        else :
            s.model = SVModelWrap(data, range(len(data)), alpha, b, **args)
    else :
        if (kernel.__class__.__name__ == 'Linear' or
            kernel.__class__.__name__ == 'Cosine') :
            s.model = LinearSVModel(data, range(len(data)), alpha, b, **args)
        else :
            s.model = SVModel(data, range(len(data)), alpha, b, **args)
    
    #s.model.svdata = data
	    
    return s

class SVR (SVM) :
    """A class for SVM regression (libsvm wrapper).
    """

    svm_type = EPSILON_SVR
    resultsObject = assess.RegressionResults
    classify = SVM.decisionFunc

    def __repr__(self) :
        rep = '<' + self.__class__.__name__ + ' instance>\n'

        return rep
    


class OneClassSVM (SVM) :
    '''wrapper for the libsvm one-class SVM'''

    svm_type = ONE_CLASS
    resultsObject = misc.DecisionFuncResults
    
    def __repr__(self) :

        rep = '<' + self.__class__.__name__ + ' instance>\n'

        return rep
                 

class OneClassCV (OneClassSVM) :

    def train(self, data, **args) :

        numFolds = 4
        oneClass = OneClassSVM(nu = self.nu)
        r = oneClass.cv(data, numFolds)
        decisionFunc = [token[1] for token in r]
        OneClassSVM.train(self, data, **args)
        offset = Numeric.sort(decisionFunc)[2]
        if offset < 0 :
            self.model.cmodel.b -= offset

        self.model.cmodel.b /= 2.0
        
class SVModel (object) :

    def __init__(self, data, svID, alpha, b, **args) :

        self.saveSpace = True
        if 'saveSpace' in args :
            self.saveSpace = args['saveSpace']

        self.alpha = alpha
        self.b = b
        self.svID = svID
        self.numSV = len(svID)
        self.svByClass = [0,0]
        # if not data.labels.numericLabels :
        #     for i in range(len(svID)) :
        #        self.svByClass[data.labels.Y[svID[i]]] += 1
        if data.X is not None :
            self.svdata = data.__class__(data, patterns = svID)
        else :
            self.svdata = None
            self.sv = svID[:]
        self.alpha = alpha[:]

    def __repr__(self) :

        rep = '<' + self.__class__.__name__ + ' instance>\n'
        rep += 'number of SVs: %d\n' % len(self)
        #for i in range(2) :
        #    rep += 'SVs in class ' + self.svdata.labels.classLabels[i] + ' : ' + \
        #           str(self.svByClass[i]) + '\n'

        return rep
    
    def __len__(self) :

        return self.numSV

    def setBias(self, bias) :

        self.b = bias

    def decisionFunc(self, x) :

        sum = 0.0
        for j in range(len(self)) :
            sum += self.svdata.kernel.eval(
                self.svdata, self.svdata.X[j], x) * self.alpha[j]

        return sum + self.b
    
    def save(self, fileName) :

        if self.saveSpace :
            raise ValueError, 'in order to save a dataset you need to train ' \
                  'as: s.train(data, saveSpace = False)'
        
	if type(fileName) == type('') :
	    outfile = open(fileName, 'w')
	else :
	    outfile = fileName

	#Classifier.save(outfile)

        outfile.write('#b=' + str(self.b) + '\n')
        outfile.write('#alpha=')
        alphaStr = [str(alpha) for alpha in self.alpha]
        outfile.write(' '.join(alphaStr))
        outfile.write('\n')
        outfile.write('#k=' + self.svdata.kernel.dump() + '\n')
        format = 'sparse'
        if self.svdata.__class__.__name__ == 'DataSet' : format = 'csv'
        self.svdata.save(outfile, format = format)

class LinearSVModel (SVModel) :

    def __init__(self, data, svID, alpha, b, **args) :

        self.saveSpace = True
        if 'saveSpace' in args :
            self.saveSpace = args['saveSpace']

        if not self.saveSpace :
            self.svdata = data.__class__(data, patterns = svID)
            self.alpha = alpha
            self.svID = svID
        self.numSV = len(svID)
        self.b = b
        self.w = self.computeW(data, svID, alpha)
        if type(self.w) == type({}) :
            self.warray = Numeric.zeros(data.numFeatures, Numeric.Float)
            for i in range(data.numFeatures) :
                if data.featureKey[i] in self.w :
                    self.warray[i] = self.w[data.featureKey[i]]
                else :
                    self.warray[i] = 0
        else :
            self.warray = self.w
        self.dotProduct = data.dotProduct
        self.svByClass = [0,0]
        if not data.labels.numericLabels :
            for i in range(len(svID)) :
                self.svByClass[data.labels.Y[svID[i]]] += 1
        if data.labels.L is not None and not data.labels.numericLabels :
            self.classLabels = data.labels.classLabels[:]

    def __repr__(self) :

        rep = '<' + self.__class__.__name__ + ' instance>\n'
        rep += 'number of SVs: %d\n' % len(self)
        if hasattr(self, 'classLabels') :
            for i in range(2) :
                rep += 'SVs in class ' + self.classLabels[i] + ' : ' + \
                       str(self.svByClass[i]) + '\n'

        return rep

    def decisionFunc(self, x) :
            
        return self.dotProduct(x, self.w) + self.b


    def computeW(self, data, svID, alpha) :


        if type(data.X[0]) == type({}) :  #Sparse dataset
            w = {}
            for i in range(len(svID)):        
                svKeys = data.X[svID[i]].keys()
                for svKey in svKeys:
                    if not w.has_key(svKey):  
                        w[svKey] = 0.0
                    w[svKey] += data.X[svID[i]][svKey] * alpha[i]
        else :                             # nonsparse dataset
            w = Numeric.zeros(len(data.featureID), Numeric.Float)
            for i in range(len(svID)):
                w += alpha[i] * data.X[svID[i]]
            
        return w


class SVModelWrap (SVModel) :

    def __init__(self, data, svID, alpha, b, **args) :

        self.saveSpace = True
        if 'saveSpace' in args :
            self.saveSpace = args['saveSpace']
        self.alpha = alpha
        self.svID = svID
        self.b = b
        self.numSV = len(svID)
        if not self.saveSpace :
            self.svdata = data.__class__(data, patterns = svID)
        self.cmodel = csvmodel.SVModel(data.castToBase(),
                                       arrayWrap.intVector(svID),
                                       arrayWrap.doubleVector(alpha), b)
    def __repr__(self) :
        rep = '<' + self.__class__.__name__ + ' instance>\n'

        return rep

    def setBias(self, bias) :

        self.b = bias
        self.cmodel.b = bias
    
    def decisionFunc (self, data, i) :

        return self.cmodel.decisionFunc(data.castToBase(), i)
        

class LinearSVModelWrap (SVModel) :

    def __init__(self, data, svID, alpha, b, **args) :

        self.saveSpace = True
        if 'saveSpace' in args :
            self.saveSpace = args['saveSpace']

        self.numSV = len(svID)

        self.cmodel = csvmodel.LinearSVModel(data, arrayWrap.intVector(svID),
                                             arrayWrap.doubleVector(alpha), b)

        self.w = self.cmodel.getWvec();
        self.warray = self.w
        self.b = b
        self.alpha = alpha
        self.svID = svID
        if not self.saveSpace :
            self.svdata = data.__class__(data, patterns = svID)

        self.numSV = len(svID)

    def setBias(self, bias) :

        self.b = bias
        self.cmodel.b = bias
        
    def decisionFunc (self, data, i) :

        return self.cmodel.decisionFunc(data, i)
        

def runMySMO(svmInstance, data) :

    C = svmInstance.getC(data)
    alphaVec = arrayWrap.doubleVector()
    b = csmo.runSMO(data.castToBase(), C, alphaVec, int(svmInstance.cacheSize))
    alpha = [alphaVec[i] for i in range(len(alphaVec))]

    return alpha,b

def runGist(classifier, data) :

    C = classifier.getC(data)
    alphaVec = arrayWrap.doubleVector()
    cgist.runGist(data.castToBase(), C, alphaVec,
                  int(classifier.cacheSize), 10000)
    alpha = [alphaVec[i] for i in range(len(alphaVec))]

    return alpha, 0.0

def runGradientDescent(classifier, data) :

    C = classifier.getC(data)
    alphaVec = arrayWrap.doubleVector()
    cgist.runGradientDescent(data.castToBase(), C, alphaVec,
                             int(classifier.cacheSize), 10000)
    alpha = [alphaVec[i] for i in range(len(alphaVec))]

    return alpha, 0.0
