#-------------------------------------------------------------------------------
# simply-predict-rt.py
#-------------------------------------------------------------------------------
# 
#  This program trains a support vector regressor (SVR) to predict
#  peptide retention time (RT). It uses as input a tab-delimited file mapping
#  peptide sequences to RTs, and then applies the trained SVR
#  to test peptides given in a file of the same format, and outputs a file
#  mapping these peptides to their actual and predicted RTs.
#  
#-------------------------------------------------------------------------------
# Author: Aaron Klammer
#-------------------------------------------------------------------------------
# $Revision: 1.3 $
#-------------------------------------------------------------------------------
# Copyright 2007, University of Washington
#-------------------------------------------------------------------------------

import getopt, sys

#-------------------------------------------------------------------------------
if __name__ == '__main__':
  if len(sys.argv) < 2: 
    print """
Usage: python %s 

  --train <filename> - File of peptides used to train SVR. Format:

                         peptide <\\t> actual RT <\\n>

  --test  <filename> - File of peptides to apply the SVR to, in above format.

  --out   <filename> - Output file containing predicted retention times. Format:

                         peptide <\\t> actual RT <\\t> predicted RT <\\n>
""" % (sys.argv[0])
    sys.exit(-1)

#-------------------------------------------------------------------------------
import Spectrum
import MatchCollection
import regressionUtils
import random
import filterTestPsm

#-------------------------------------------------------------------------------
if __name__ == '__main__':
  optlist, args = getopt.getopt(sys.argv[1:],'', ["train=", "test=", "out="])

  trainFname = testFname = outFname = None 

  #--- process options
  for option_tuple in optlist:
    option_name = option_tuple[0]; option_value = option_tuple[1];
    if option_name== '--train' : trainFname = option_value
    if option_name== '--test'  : testFname  = option_value
    if option_name== '--out'   : outFname   = option_value

  #--- set default parameters for regression
  svrParams = {}
  svrParams["kernel"] = "gaussian"
  svrParams["wpowers"] = [-6.0, -7.0, -8.0]
  svrParams["cpowers"] = [5.0, 6.0, 7.0]

  features = {}
  features["twomer"] = features["tryptic"] = False
  features["nterm"] = features["cterm"] = True

  #--- read in training file
  print >>sys.stderr, "Reading training file %s" % trainFname
  trainPsms = MatchCollection.PsmMatchCollection()
  trainFh = open(trainFname)
  for line in trainFh:
    fields = line.rstrip("\n").split()
    peptide = fields[0]
    rt = float(fields[1])
    spectrum = Spectrum.Ms2Spectrum()
    spectrum["retention time"] = rt
    trainPsms.addPsmsForSpectrum(spectrum, [peptide])
  trainFh.close()

  #--- read in test file
  print >>sys.stderr, "Reading test file %s" % testFname
  testPepRt = {}
  testFh = open(testFname)
  for line in testFh:
    fields = line.rstrip("\n").split()
    testPepRt[fields[0]] = float(fields[1])
  testFh.close()
  
  #--- train the SVR for predicting RT
  print >>sys.stderr, "Performing regression"
  random.shuffle(trainPsms["matches"])
  bestSvrArgs = regressionUtils.trainFoldsSVR(trainPsms["matches"], 
    features=features, svrParams=svrParams)

  #--- apply best params for the SVR to the test peptides
  print >>sys.stderr, "Applying regression parameters to test peptides"
  peptides = testPepRt.keys()
  results = regressionUtils.trainTestPeptides(trainPsms["matches"], 
      peptides, bestSvrArgs, features=features)

  #--- output predicted retention times
  print >>sys.stderr, "Outputting to %s" % outFname
  outFh = open(outFname, "w")
  for idx in range(len(peptides)): 
    print >>outFh, "%s\t%.3f\t%.3f" \
      % (peptides[idx], testPepRt[peptides[idx]], float(results[idx]))
  outFh.close()
 
  print >>sys.stderr, "Finished processing" 
